# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/07_misc.ipynb.

# %% auto 0
__all__ = ['get_option', 'find_spec_or_none', 'find_last_spec', 'getretann', 'prep_kwargs', 'asset_attrs', 'asset_sub_pkg',
           'asset_last', 'asset_name', 'parse_asset', 'insys', 'find_asset_module', 'load_asset', 'try_loaded_attr',
           'try_ducked_attr', 'unduck', 'hatchegg', 'putinrow', 'prepare_ducktypes', 'prepare_isinstance_types',
           'prepare_types_for_instancecheck', 'tryattr']

# %% ../nbs/07_misc.ipynb 6
import io, sys
from inspect import get_annotations
from functools import wraps

from contextlib import redirect_stderr, redirect_stdout

from importlib.abc import Loader
from importlib.util import (module_from_spec, find_spec)
from importlib.machinery import ModuleSpec

# %% ../nbs/07_misc.ipynb 8
from types import ModuleType
from typing import Any, Union, Iterable, Optional, get_args

# %% ../nbs/07_misc.ipynb 10
#| export

# %% ../nbs/07_misc.ipynb 12
#| export

# %% ../nbs/07_misc.ipynb 14
from quac.cons import (
    DOT, NIL, LOAD, LOADER, RETURN, DUCK_SPEC_ATTRS, 
    __DOC__, __NEW__, __CALL__, __LOAD__, __INIT__, __NAME__, __MODULE__, __SIGNATURE__, __ANNOTATIONS__,
    __HATCHED__, __ALLCHICS__, __AREINROW__, __DUCKTYPE__
)
from .type import T, P
from .prot import NotSpecial, DuckSpecProtocol, DuckModuleProtocol, DuckSpecProtocol, DuckTypeProtocol
from .grds import iscls, isnotspecial, isnotiterstr, isduck, isduckspec, isduckmodule, hasduck, hasload, isclstype, _istype, isoptional, hasducktype
from .errs import (DuckException, DuckNameError, DuckImportError, DuckAttributeError, DuckModuleNotFoundError)

# %% ../nbs/07_misc.ipynb 17
def get_option(x, i: int = 0):
    "Check if `x` is an optional type"
    if not isoptional(x): return None
    args = get_args(x)
    if len(args) == 0: return None
    return args[i if i < len(args) else -1] 

# %% ../nbs/07_misc.ipynb 19
@wraps(find_spec)
def find_spec_or_none(name: str, package: str | None = None) -> ModuleSpec | None:
    # NOTE: needed as find_spec will fail if stub is an attribute e.g. 
    # `pandas.core.arrays.categorical.CategoricalAccessor`
    try: return find_spec(name, package)
    except ModuleNotFoundError: return None

def find_last_spec(fullpath: str) -> tuple[ModuleSpec | None, str | None]:
    stubs = fullpath.split(DOT)
    while len(stubs) > 0:
        stub = DOT.join(stubs)
        spec = find_spec_or_none(stub)
        if spec: break
        stubs.pop()
    return spec, stub

# %% ../nbs/07_misc.ipynb 20
def getretann(obj) -> Any:
    if hasattr(obj, __ANNOTATIONS__): return get_annotations(obj).get(RETURN, None)
    else: return None

# %% ../nbs/07_misc.ipynb 22
def prep_kwargs(obj: object, *args: P.args, __attrs: tuple[str, ...] = DUCK_SPEC_ATTRS, **kwargs: P.kwargs) -> dict:
    objkws, argkws = dict(),  dict(zip(__attrs, args))
    for i, attr in enumerate(__attrs):
        objkws.setdefault(attr, getattr(obj, attr, None))
        argkws.setdefault(attr, objkws.get(attr))
        kwargs.setdefault(attr, argkws.get(attr))
    return kwargs

# %% ../nbs/07_misc.ipynb 24
def asset_attrs(asset) -> tuple[str, ...]:
    *_, rest = asset.removesuffix(DOT).partition(DOT)
    return rest.split(DOT)

def asset_sub_pkg(asset) -> tuple[str, str]:
    name, _, attr = asset.removesuffix(DOT).rpartition(DOT)
    return name, attr

def asset_last(asset: str) -> str:
    *_, name = str(asset).removesuffix(DOT).rpartition(DOT)
    return name

def asset_name(asset, is_module: bool = False) -> str:
    if is_module or DOT not in asset: return asset
    name, _ = asset_sub_pkg(asset)
    return name

def parse_asset(asset: str, is_module: bool = False) -> tuple[str, str | None]:
    if is_module or DOT not in asset: return asset, None
    name, attr = asset_sub_pkg(asset)
    return name, attr

# %% ../nbs/07_misc.ipynb 29
def insys(asset: str, is_module: bool = False) -> bool:
    return asset_name(asset, is_module) in sys.modules

def find_asset_module(asset: str, is_module: bool = False) -> ModuleType:
    name, attr = parse_asset(asset, is_module)
    if insys(asset, is_module): return sys.modules[name]
    
    spec: ModuleSpec | None = find_spec(name)
    module: ModuleType = module_from_spec(spec)
    loader: Loader = getattr(spec, LOADER, None)
    
    try: loader.exec_module(module)
    except NameError as e: ...
    return module

def load_asset(asset: str, is_module: bool = False):
    name, attr = parse_asset(asset, is_module)
    module = find_asset_module(asset, is_module)
    if is_module or attr is None: return module
    return getattr(module, attr)

def try_loaded_attr(asset: str, is_module: bool = False, hush: bool = False):
    if not hush: return load_asset(asset, is_module)
    with redirect_stderr(io.StringIO()), redirect_stdout(io.StringIO()):
        return load_asset(asset, is_module)

# %% ../nbs/07_misc.ipynb 33
def try_ducked_attr(asset: str, ducked: Optional[DuckSpecProtocol] = None):
    if ducked is None: raise DuckImportError(asset)
    try:
        item = ducked
        for attr in asset_attrs(asset):
            item = getattr(item, attr)
        return item
    except NameError as e: 
        raise DuckNameError(*e.args)
    
    except ModuleNotFoundError as e: 
        raise DuckModuleNotFoundError(*e.args, e.name, e.path)
    
    except ImportError as e: 
        raise DuckImportError(*e.args, e.name, e.path)
    
    except AttributeError as e: 
        raise DuckAttributeError(*e.args, e.name, e.obj)
    
    except Exception as e: 
        raise DuckException(*e.args)
    
    return item

# %% ../nbs/07_misc.ipynb 35
def unduck(item: Any):
    try:
        if isduckspec(item): return item.duck
        elif callable(getattr(item, LOAD, None)):
            return item.load()
        elif callable(getattr(item, __LOAD__, None)):
            return item.__load__()
    except Exception: ...
    return item

def _fromducks(var: Iterable):
    iterstr = isnotiterstr(var)
    if not iterstr: return var
    cls = type(var)
    var = list(var)
    for i, e in enumerate(var): 
        var[i] = unduck(e)
    return cls(var)

def _fromcall(var: Iterable):
    if not callable(var): 
        return var
    try: return var()
    except: ...
    return var

# %% ../nbs/07_misc.ipynb 36
def hatchegg(cls: DuckSpecProtocol) -> tuple[type, bool, bool]:
    '''Try to load the asset stored in the duck specification.
    
    Returns
    -------
        chic : type, optional
            the loaded asset from a `DuckSpec`
        
        eggq : bool
            Whether the input `cls` was a duckspec
        
        isok : bool
            Whether the attempt to load the asset succeeded or not.
    '''
    if not (isduckspec(cls) or hasduck(cls)): 
        return None, False, True
    try:
        if not getattr(cls, __HATCHED__, False):
            chic = cls.duck
            setattr(cls, __HATCHED__, True)
            return chic, True, True
        else:
            return cls.duck, True, True
    except Exception as e: ...
    return None, True, False
    


# %% ../nbs/07_misc.ipynb 38
def putinrow(cls: DuckTypeProtocol, force: bool = False):
    '''
    Returns
    -------
    types: tuple[type, ...]
        The resolved types of the given ducktype. Note that if `cls`
        is not a duck, then returns `(cls, )`
    
    inrow: bool
        Whether all of the types in `cls.__ducktype__` are resolved.
        If `cls` is not a duck, then defaults to True
        
    duckq: bool
        Whether or not the input class `cls` was a ducktype to begin with.
    '''
    if not (isduck(cls) or hasducktype(cls)):
        return (cls, ), True, False
    
    # print('putinrow, cls=', cls.__name__)
    curtypes = list(cls.__ducktype__)
    
    if getattr(cls, __AREINROW__, False) and not force: 
        return tuple(curtypes), True, True
    
    gathered = dict() # collection of found types, to prevent inf recursion through instance checking
    areinrow = [] # whether each type of __ducktype__ is "in row"
    foundrow = [] # order in which types were resolved
    
    # for each type in Duck.__ducktype__
    for t in curtypes:
        # if t is a DuckSpec (`egg`) load the item (`chic`)
        chic, eggq, isok = hatchegg(t)
        if eggq:
            # Add the resolved `DuckSpec`'s item to gathered and put in the foundrow
            # if isok otherwise keep the original `DuckSpec` in the ducktype
            use = chic if isok else t
            if not gathered.get(use, False):
                gathered[use] = True
                foundrow.append(use)
            
            # keep track of if this type was resolved or not
            areinrow.append(isok)
            continue             
        
        # if t is also a Duck, we need to recurse
        if (isduck(t) or hasducktype(t)):
            inrow = getattr(t, __AREINROW__, False) and not force
            if not inrow:
                # attempt to put `t.__ducktype__` in row.
                try:
                    types, inrow, duckq = putinrow(t, force)
                    setattr(t, __AREINROW__, inrow)
                    areinrow.append(inrow)
                    setattr(t, __DUCKTYPE__, tuple(types))
                except:
                    # failed to get them in row
                    setattr(t, __AREINROW__, False)
                    areinrow.append(False)
            else:
                # they are already in row
                areinrow.append(True)
            
            # add each or `t.__ducktype__`'s to gathered and 
            # put in the foundrow if missing
            subducks = list(t.__ducktype__)
            for d in subducks:
                if not gathered.get(d, False):
                    gathered[d] = True
                    foundrow.append(d)
                    continue
                
        else:
            # not a duck or a duckspec, so they are fine as is
            if not gathered.get(t, False):
                gathered[t] = True
                foundrow.append(t)
            areinrow.append([True])
    
    # determine if this operation completed successfully
    areinrow = all(areinrow)
    setattr(cls, __AREINROW__, areinrow)
    
    foundrow = tuple(foundrow) # __ducktype__ is a tuple
    setattr(cls, __DUCKTYPE__, foundrow)
    return foundrow, areinrow, True

# %% ../nbs/07_misc.ipynb 42
def prepare_ducktypes(cls, force: bool = False) -> tuple[type, ...]:
    unducked, *_ = putinrow(cls, force=force)
    return unducked
    
    curtypes = list(cls.__ducktype__)
    unducked = []
    found = dict()
    for t in curtypes:
        if isduckspec(t) or hasduck(t):
            # print('isduckspec')
            try:
                t = t.duck
                if not found.get(t, False):
                    found[t] = True
                    unducked.append(t)
                    continue
            except Exception as e: ...
                
        if isduck(t):    
            # print('isduck')
            resolved = list(t.__ducktype__)
            for d in resolved:
                if not found.get(d, False):
                    found[d] = True
                    unducked.append(d)
                    continue
            # unducked.extend(list(t.__types__()))
            continue
        
        if not found.get(t, False):
            found[t] = True
            unducked.append(t)
            continue
        # unducked.append(t)
            
            
    # unducked = list(set(unducked))
    
    # unducked = list(filter(lambda e: unducked.count(e) == 1, unducked))    
    # unducked = sorted(
    #     unducked, 
    #     key = lambda t: (
    #         iscls(t), 
    #         getattr(t, '__name__', ''),
    #         curtypes.index(t) if t in curtypes else -1,
    #     ),
    #     reverse=True
    # )
    
    unducked = tuple(unducked)
    return unducked

# %% ../nbs/07_misc.ipynb 43
def prepare_isinstance_types(types: tuple[type | Any, ...]) -> tuple[type, ...]:
    types = tuple([t if _istype(t) else type(t) for t in types])
    return types

# %% ../nbs/07_misc.ipynb 44
def prepare_types_for_instancecheck(types: tuple[type, ...]) -> tuple[type, ...]:
    results = []
    for t in types:
        if hasduck(t): 
            try:
                results.append(t.duck)
                continue
            except Exception: ...
            
        if hasload(t): 
            try:
                results.append(t.load())
                continue
            except Exception: ...
            
        if isclstype(t): 
            results.append(t)
            
        else: 
            results.append(type(t))
    
    return tuple(results)

# %% ../nbs/07_misc.ipynb 46
def tryattr(
    asset: str, 
    is_module: bool = False, 
    ducked: Optional[DuckModuleProtocol] = None,
    default: Any = None,
    bases: tuple = (),
    attrs: dict = {},
    hush: bool = True,
    delayed: dict[str, bool] = dict(),
    **kwargs
) -> type:
    item, ierr, derr = None, False, False
    try: item = try_loaded_attr(asset, is_module, hush=hush)         
    except (NameError, ImportError, AttributeError): ierr = True    
    
    if isinstance(item, ModuleType) and is_module: return item
    elif not ierr: return item
    
    try: item = try_ducked_attr(asset, ducked)
    except (DuckException, DuckNameError, DuckImportError, DuckAttributeError, DuckModuleNotFoundError): derr = True
            
    attr = asset_last(asset)
    
    isducked = isduckmodule(item)


    itemname = getattr(item, __NAME__, None)
    goodname, notfound = attr == itemname, attr != itemname
    maketype: bool = (isducked and notfound) or (ierr or derr) or default in (None, type)
    
    found = not isducked and goodname
    if found: return item
    
    
    for k, v in delayed.items():
        if not v and k in locals(): continue
        var = locals().get(k)
        var = _fromducks(var)
        var = _fromcall(var)
        try:
            match k:
                case 'attrs': attrs = var()
                case 'bases': bases = var()
                case 'default': default = var()
                case _: ...
        except: ...
    
    try:
        if maketype: 
            # moddict = dict(__module__=globals().get(__MODULE__, 'quac'))
            # attrs.update(moddict)
            item = type(attr, bases, attrs)
            
        else: item = default
    
    except Exception: 
        item = default
    
    item = unduck(item)
    return item



# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/40_duck.ipynb.

# %% auto 0
__all__ = ['iterable', 'itertype', 'iterq', 'listq', 'tupleq', 'nlen', 'len1', 'len2', 'len3', 'len4', 'npbool', 'boolic',
           'bool01', 'bools', 'boolq', 'intq', 'ints', 'floatq', 'floats', 'real', 'realq', 'reals', 'num', 'numq',
           'nums', 'numstr', 'numstrq', 'none', 'notnone', 'na', 'anyq', 'pynan', 'pdna', 'npnan', 'nan', 'nanq',
           'nans', 'pattern', 'rematch', 'strtmpl', 'patternq', 'nilstr', 'notnilstr', 'strbytes', 'notstrbytes',
           'notiterstr', 'strq', 'strs', 'ospathlike', 'plibpath', 'posixpath', 'pathlike', 'pathq', 'paths',
           'walktuple', 'walkparts', 'moduleloader', 'modulespec', 'moduletype', 'wrappertype', 'builtinfunctype',
           'trackbacktype', 'functiontype', 'mapping', 'moduleq', 'callq', 'literal', 'enumauto', 'enumdict',
           'dictkeys', 'dictvalues', 'dictitems', 'setq', 'dictq', 'sig', 'param', 'paramkind', 'logger', 'loggerq',
           'ichain', 'iprod', 'ziplong', 'artist', 'colormap', 'lcolormap', 'lscolormap', 'mplpatch', 'mplaxbase',
           'mplaxes', 'nparray', 'npmatrix', 'recarray', 'npempty', 'npdatasource', 'sparray', 'spmatrix', 'cscarray',
           'cscmatrix', 'cooarray', 'coomatrix', 'csrarray', 'csrmatrix', 'diaarray', 'diamatrix', 'dokarray',
           'dokmatrix', 'lilarray', 'lilmatrix', 'bsrarray', 'bsrmatrix', 'sparsearray', 'sparsematrix', 'sparse',
           'array', 'matrix', 'sparseq', 'arrayq', 'matrixq', 'cv2vidcap', 'pilimage', 'vidcap', 'image', 'groupkey',
           'pddelegate', 'pdcataccessor', 'pdindex', 'pdcatindex', 'pdseries', 'boolindex', 'catindex', 'boolindexq',
           'catindexq', 'indextype', 'indexlike', 'seriestype', 'serieslike', 'pddataframe', 'dataframe', 'pttype',
           'pttypeq', 'memfmt', 'presfmt', 'ptsize', 'ptsizeq', 'ptdevice', 'ptdeviceq', 'pttensor', 'inttensor',
           'booltensor', 'floattensor', 'longtensor', 'doubletensor', 'pttensorq', 'inttensorq', 'booltensorq',
           'floattensorq', 'longtensorq', 'doubletensorq', 'nantensor', 'nantensorq', 'tensor', 'tensorq', 'ptmodule',
           'plmodule', 'nnmodule', 'ptdataset', 'pldatamodule', 'ptdataloader', 'neuralode', 'neuralsde',
           'neuraldynamics', 'augmenter', 'odefunc', 'sdefunc', 'sdeito', 'brownianintervial', 'dtype', 'dtypeq',
           'device', 'deviceq', 'devcpu', 'devmps', 'devcuda', 'size', 'sizeq', 'datasource', 'adata', 'adatas',
           'graphtype', 'datatype', 'nnum', 'nint', 'nfloat', 'int2', 'int3', 'int4', 'shaped', 'dimmed', 'dim1',
           'dim2', 'dim3', 'dim4', 'genset', 'setstr', 'setnum', 'kvdict', 'pttndict', 'annreturn', 'slicelike',
           'slicespec', 'slicespecs', 'richtree', 'richtext', 'richcons', 'richprog', 'richtask', 'tqdmauto',
           'typerapp']

# %% ../nbs/40_duck.ipynb 6
from enum import member, _EnumDict

# %% ../nbs/40_duck.ipynb 8
from types import FunctionType
from typing import (
    Any, Dict, List, Set, Self, Type, Union, Tuple, Generic, Callable, Literal, 
    TypeVar, ClassVar, TypeAlias, ParamSpec, Optional,
    Iterable, ForwardRef, Mapping, TypeGuard, Generic, Protocol
)

# %% ../nbs/40_duck.ipynb 10
#| export

# %% ../nbs/40_duck.ipynb 12
#| export

# %% ../nbs/40_duck.ipynb 14
from quac.cons import (
    NIL, NDIM, SHAPE, KEYS, VALUES, __K__, __V__, __N__, __RETANNOT__, __KVTYPES__,
    __D__, __LIT__, __LITNAME__, __EMPTYVAL__, __DEV__
)
from .type import T, Ts, O, P
from .enum import Species
from .base import Duck
from .misc import getretann
from . import grds, eggs

# %% ../nbs/40_duck.ipynb 17
class iterable(Duck, Generic[*Ts]):
    '''`typings.Iterable`, `collections.abc.Iterable`'''
    __ducktype__ = (Iterable, eggs.c_iterable, )
    __generate__ = lambda: ()
    
    
class itertype(iterable, Generic[*Ts]):
    '''iterable type (i.e. `Iterable`, `list`, or `tuple`)'''
    __ducktype__ = (iterable, list, tuple)
    __generate__ = lambda: ()

# %% ../nbs/40_duck.ipynb 18
class iterq(itertype, Generic[*Ts]):
    __species__ = Species.OPT

# %% ../nbs/40_duck.ipynb 21
class listq(Duck[list[T]]):
    __ducktype__ = (list, )
    __species__ = Species.OPT

class tupleq(Duck[tuple]):
    __ducktype__ = (tuple, )
    __species__ = Species.OPT

# %% ../nbs/40_duck.ipynb 24
from typing import Any
class nlen(itertype[*Ts], Duck):
    '''A ducktype for iterables with a specific length'''
    __n__: ClassVar[int] = 0
    __emptyval__: ClassVar[Any] = 0
    
    @classmethod
    def __metahook__(cls, mcls, name, bases, namespace, **kwargs):
        super().__metahook__(mcls, name, bases, namespace, **kwargs)
        n = cls.__n__
        e = cls.__emptyval__
        islen = grds._islen(cls)
        cls.__expected__.update(__len__=n)
        cls.__extrafns__.update(__len__=islen)        
        cls.__generate__ = lambda: (e, ) * n
        return cls

# %% ../nbs/40_duck.ipynb 25
class len1(nlen[*Ts]):
    '''A ducktype for iterables with length 1'''
    __n__: ClassVar[int] = 1

class len2(nlen[*Ts]):
    '''A ducktype for iterables with length 2'''
    __n__: ClassVar[int] = 2

class len3(nlen[*Ts]):
    '''A ducktype for iterables with length 3'''
    __n__: ClassVar[int] = 3

class len4(nlen[*Ts]):
    '''A ducktype for iterables with length 4'''
    __n__: ClassVar[int] = 4

# %% ../nbs/40_duck.ipynb 30
class npbool(Duck): 
    '''`np.bool_`'''
    __ducktype__ = (eggs.np_bool, )

class boolic(npbool): 
    '''`bool`, `np.bool_`'''
    __ducktype__ = (eggs.np_bool, bool,)

class bool01(Duck):
    '''`bool`, `np.bool_`, `0`, `1`'''
    __ducktype__ = (boolic, 0, 1)
    
class bools(itertype[*Ts]):
    '''list of bool01'''
    __expected__ = dict(dtype=bool01)

# %% ../nbs/40_duck.ipynb 31
class boolq(bool01):
    __species__ = Species.OPT

# %% ../nbs/40_duck.ipynb 35
class intq(Duck):
    '''`int` or None'''
    __ducktype__ = (int, )
    __species__ = Species.OPT
    
class ints(itertype[*Ts], Generic[*Ts]):
    '''iterable of `int`s'''
    __expected__ = dict(dtype=int)

# %% ../nbs/40_duck.ipynb 37
class floatq(Duck):
    '''`float` or None'''
    __ducktype__ = (float, )
    __species__ = Species.OPT
    
class floats(itertype[*Ts], Generic[*Ts]):
    '''iterable of `float`s'''
    __expected__ = dict(dtype=float)

# %% ../nbs/40_duck.ipynb 39
class real(Duck):
    '''R (i.e. `int` or `float`)'''
    __ducktype__ = (int, float, )
    
class realq(real):
    '''`real` or None'''
    __species__ = Species.OPT
    
class reals(itertype[*Ts], Generic[*Ts]):
    '''iterable of `real`'''
    __expected__ = dict(dtype=real)

# %% ../nbs/40_duck.ipynb 41
class num(Duck):
    '''`int`, `float`, `complex`, or `Number`'''
    __ducktype__ = (int, float, complex, eggs.n_number)
    
class numq(num):
    '''`real` or None'''
    __species__ = Species.OPT
    
class nums(itertype[*Ts], Generic[*Ts]):
    '''iterable of `num`'''
    __expected__ = dict(dtype=num)

# %% ../nbs/40_duck.ipynb 45
class numstr(Duck):
    '''`num` or str'''
    __ducktype__ = (num, str,)

# %% ../nbs/40_duck.ipynb 46
class numstrq(numstr):
    __species__ = Species.OPT

# %% ../nbs/40_duck.ipynb 49
class none(Duck): 
    '''`None` or `NoneType`'''
    __ducktype__ = (None, eggs.t_nonetype)
    
class notnone(none): 
    '''not `None` or `NoneType`'''
    __species__ = Species.NOT
    
class na(Duck): 
    '''`NA` or `NAType`'''
    __ducktype__ = (eggs.pd_na, eggs.pd_natype,)

# %% ../nbs/40_duck.ipynb 50
class anyq(notnone[T]):
    __species__ = Species.OPT

# %% ../nbs/40_duck.ipynb 52
class pynan(Duck): 
    '''`math.nan`''';
    __ducktype__ = (eggs.m_nan, )
    
class pdna(Duck): 
    '''`pd.NA`''';
    __ducktype__ = (
        eggs.pd_na, 
    )

class npnan(Duck): 
    '''`np.nan`, `np.NaN`, or `np.NAN`'''
    __ducktype__ = (
        eggs.np_nan, eggs.np_NaN, eggs.np_NAN, 
    )

# %% ../nbs/40_duck.ipynb 53
class nan(Duck):
    '''`math.nan`, `pd.NA`, `np.nan`, `np.NaN`, or `np.NAN`'''
    __ducktype__ = (
        eggs.m_nan, eggs.np_nan, eggs.np_NaN, eggs.np_NAN, 
        eggs.pd_na,# eggs.pd_natype, #float('nan')
    )
    
class nanq(Duck):
    '''`nan` or None'''
    __ducktype__ = (nan, )
    __species__ = Species.OPT
    
class nans(Duck):
    '''iterable of `nan`'''
    __ducktype__ = (itertype,)
    __expected__ = dict(dtype=nan)

# %% ../nbs/40_duck.ipynb 60
class pattern(Duck): 
    '''`re.Pattern`'''
    __ducktype__ = (eggs.re_pattern, )

class rematch(Duck): 
    '''`re.Match`'''
    __ducktype__ = (eggs.re_match, )

class strtmpl(Duck): 
    '''`string.Template`'''
    __ducktype__ = (eggs.s_template, )

# %% ../nbs/40_duck.ipynb 61
class patternq(pattern):
    __species__ = Species.OPT

# %% ../nbs/40_duck.ipynb 63
class nilstr(Duck):
    '''`''` (the nil string)'''
    __ducktype__ = (NIL, )
    __guard__ = lambda x: x == NIL
    
class notnilstr(nilstr): 
    '''not `nilstr` (not `''`, the nil string)'''
    __species__ = Species.NOT

# %% ../nbs/40_duck.ipynb 64
class strbytes(Duck):
    '''str or bytes'''
    __ducktype__ = (str, bytes)
    
class notstrbytes(strbytes): 
    '''not (str or bytes)'''
    __species__ = Species.NOT
    

class notiterstr(itertype[*Ts], Generic[*Ts]):
    '''An iterable but not of `strbytes`'''
    # __ducktype__ = (itertype, )
    @classmethod
    def __guard__(cls, x):
        return super().__guard__(x) and notstrbytes(x)

# %% ../nbs/40_duck.ipynb 65
class strq(Duck): 
    '''`str` or None'''
    __ducktype__ = (str, )
    __species__ = Species.OPT

class strs(Duck): 
    '''iterable of `str`'''
    __ducktype__ = (itertype, notstrbytes, )
    __expected__ = dict(dtype=str)

# %% ../nbs/40_duck.ipynb 68
class ospathlike(Duck):
    '''`os.PathLike`'''
    __ducktype__ = (eggs.os_pathlike, )

class plibpath(Duck):
    '''`pathlib.Path`'''
    __ducktype__ = (eggs.pl_path, )

class posixpath(Duck):
    '''`pathlib.PosixPath`'''
    __ducktype__ = (eggs.pl_posixpath, )

# %% ../nbs/40_duck.ipynb 70
class pathlike(Duck, Generic[*Ts]):
    '''`os.PathLike`, `pathlib.Path`, `pathlib.PosixPath`'''
    __ducktype__ = (str, bytes, eggs.os_pathlike, eggs.pl_path, eggs.pl_posixpath)    

# %% ../nbs/40_duck.ipynb 71
class pathq(pathlike, Generic[*Ts]):
    __species__ = Species.OPT

# %% ../nbs/40_duck.ipynb 73
class paths(notiterstr):
    '''A type alias for a list of paths.'''
    __expected__ = dict(dtype=pathlike)

# %% ../nbs/40_duck.ipynb 75
class walktuple(len3, itertype[pathlike, paths, paths]):
    '''A type alias for a tuple from `os.walkdir` i.e. `(dirpath, dirnames, filenames)`.'''
    __expected__ = dict(dtype=pathlike)
    @classmethod
    def __guard__(cls, obj: Any) -> bool:
        return grds.istuple(obj) and len(obj) == 3 and pathlike(obj[0]) and all(paths(*obj[1:]))
    
class walkparts(itertype[walktuple]):
    '''A type alias for an iterable of walk tuples from `os.walkdir` 
    i.e. `Iterable[(dirpath, dirnames, filenames), ...]`.'''
    __expected__ = dict(dtype=walktuple)

# %% ../nbs/40_duck.ipynb 77
class moduleloader(Duck):
    '''`importlib.Loader`'''
    __ducktype__ = (eggs.il_moduleloader, )
    
class modulespec(Duck):
    '''`importlib.machinery.ModuleSpec`'''
    __ducktype__ = (eggs.il_modulespec, )

# %% ../nbs/40_duck.ipynb 79
class moduletype(Duck):
    '''`types.ModuleType`'''
    __ducktype__ = (eggs.t_moduletype, )
    
class wrappertype(Duck):
    '''`types.MethodWrapperType`'''
    __ducktype__ = (eggs.t_methodwrappertype, )
    
class builtinfunctype(Duck):
    '''`types.BuiltinFunctionType`'''
    __ducktype__ = (eggs.t_builtinfunctiontype, )
    
class trackbacktype(Duck):
    '''`types.TracebackType`'''
    __ducktype__ = (eggs.t_tracebacktype, )
    
class functiontype(Duck):
    '''`types.FunctionType`'''
    __ducktype__ = (eggs.t_functiontype, )
    
class mapping(Duck):
    '''`collections.Mapping`'''
    __ducktype__ = (eggs.c_mapping, )
    

# %% ../nbs/40_duck.ipynb 80
class moduleq(moduletype, Generic[T]):
    '''`Optional[types.ModuleType]`'''
    __species__ = Species.OPT

# %% ../nbs/40_duck.ipynb 82
class callq(Duck[Callable[P, T]]):
    '''`Optional[Callable[P, T]]`'''
    __ducktype__ = (Callable, )
    __species__ = Species.OPT

# %% ../nbs/40_duck.ipynb 84
class literal(Duck[T]):
    '''`typing.Literal`'''
    __ducktype__ = (Literal, str, )
    __lit__: ClassVar[Optional[T]] = None
    __litname__: ClassVar[bool] = True
    @classmethod
    def __metahook__(cls, mcls, name, bases, namespace, **kwargs):
        __lit__ = kwargs.get(__LIT__, getattr(cls, __LIT__, None))
        __name__ = f'[{__lit__}]'
        rename = getattr(cls, __LITNAME__, True)
        namespace.update(__lit__=__lit__)
        super().__metahook__(mcls, name, bases, namespace, **kwargs)
        setattr(cls, __LIT__, __lit__)
        isval = lambda obj, val: val == __lit__
        cls.__expected__.update(__lit__=__lit__)
        cls.__extrafns__.update(__lit__=isval)
        if rename: cls.__name__ = __name__
        return cls

# %% ../nbs/40_duck.ipynb 86
class enumauto(Duck[member]):
    '''A type alias for an `auto()` generated enum `member`.'''
    __ducktype__ = (member, )

class enumdict(Duck[_EnumDict]):
    '''A type alias for `_EnumDict`.'''
    __ducktype__ = (_EnumDict, )

# %% ../nbs/40_duck.ipynb 88
class dictkeys(Duck):
    '''`_collections_abc.dict_keys`'''
    __ducktype__ = (eggs.c_dict_keys, )
    
class dictvalues(Duck):
    '''`_collections_abc.dict_values`'''
    __ducktype__ = (eggs.c_dict_values, )
    
class dictitems(Duck):
    '''`_collections_abc.dict_items`'''
    __ducktype__ = (eggs.c_dict_items, )

# %% ../nbs/40_duck.ipynb 89
class setq(Duck[set]):
    '''`Optional[set]`'''
    __ducktype__ = (set, )
    __species__ = Species.OPT
    
class dictq(Duck[dict[T, O]]):
    '''`Optional[dict]`'''
    __ducktype__ = (dict, )
    __species__ = Species.OPT

# %% ../nbs/40_duck.ipynb 91
class sig(Duck):
    '''`inspect.Signature`'''
    __ducktype__ = (eggs.i_sig, )
    
class param(Duck):
    '''`inspect.Parameter`'''
    __ducktype__ = (eggs.i_param, )
    
class paramkind(Duck):
    '''`inspect._ParameterKind`'''
    __ducktype__ = (eggs.i_paramkind, )

# %% ../nbs/40_duck.ipynb 93
class logger(Duck):
    '''`logging.Logger`'''
    __ducktype__ = (eggs.l_logger, )
    
    
class loggerq(logger):
    '''`Optional[logging.Logger]`'''
    __species__ = Species.OPT

# %% ../nbs/40_duck.ipynb 95
class ichain(Duck):
    '''`itertools.chain`'''
    __ducktype__ = (eggs.it_chain, )
    
class iprod(Duck):
    '''`itertools.product`'''
    __ducktype__ = (eggs.it_prod, )
    
class ziplong(Duck):
    '''`itertools.zip_longest`'''
    __ducktype__ = (eggs.it_ziplong, )

# %% ../nbs/40_duck.ipynb 98
class _mpl_base(Duck):
    '''`matplotlib.axes._base`''';
    __ducktype__ = (eggs._mpl_base, )
    
class _mpl_axes(Duck):
    '''`matplotlib.axes._axes`''';
    __ducktype__ = (eggs._mpl_axes, )
    

class artist(Duck): 
    '''`matplotlib.artist.Artist`'''
    __ducktype__ = (eggs.mpl_artist, )
    
class colormap(Duck): 
    '''`matplotlib.colors.Colormap`'''
    __ducktype__ = (eggs.mpl_colormap, )

class lcolormap(Duck): 
    '''`matplotlib.colors.ListedColormap`'''
    __ducktype__ = (eggs.mpl_lcolormap, )
    
class lscolormap(Duck): 
    '''`matplotlib.colors.LinearSegmentedColormap`'''
    __ducktype__ = (eggs.mpl_lscolormap, )
    
class mplpatch(Duck): 
    '''`matplotlib.patches.Patch`'''
    __ducktype__ = (eggs.mpl_patch, )

class mplaxbase(Duck): 
    '''`matplotlib.axes._base._AxesBase`'''
    __ducktype__ = (eggs.mpl_axbase, )

class mplaxes(Duck): 
    '''`matplotlib.axes._axes.Axes`'''
    __ducktype__ = (eggs.mpl_axes, )
    

# %% ../nbs/40_duck.ipynb 101
class nparray(Duck, Generic[*Ts]): 
    '''`numpy.ndarray`''';
    __ducktype__ = (eggs.np_array, )
    
class npmatrix(Duck, Generic[*Ts]): 
    '''`numpy.matrix`''';
    __ducktype__ = (eggs.np_matrix, )

class recarray(Duck, Generic[*Ts]): 
    '''`numpy.recarray`''';
    __ducktype__ = (eggs.np_recarray, )
    
class npempty(Duck):
    '''`numpy.empty`''';
    __ducktype__ = (eggs.np_empty, )

# %% ../nbs/40_duck.ipynb 102
class npdatasource(Duck): 
    '''`numpy.DataSource`''';
    __ducktype__ = (eggs.np_datasource, )

# %% ../nbs/40_duck.ipynb 104
class sparray(Duck, Generic[*Ts]): 
    '''`scipy.sparse.aparray`''';
    __ducktype__ = (eggs.sp_sparray, )

class spmatrix(Duck, Generic[*Ts]): 
    '''`scipy.sparse.spmatrix`''';
    __ducktype__ = (eggs.sp_spmatrix, )
    

# %% ../nbs/40_duck.ipynb 106
class cscarray(Duck):
    '''`scipy.sparse.csc_array`''';
    __ducktype__ = (eggs.sp_csc_array, )
    
class cscmatrix(Duck):
    '''`scipy.sparse.csc_matrix`''';
    __ducktype__ = (eggs.sp_csc_matrix, )

class cooarray(Duck):
    '''`scipy.sparse.coo_array`''';
    __ducktype__ = (eggs.sp_coo_array, )

class coomatrix(Duck):
    '''`scipy.sparse.coo_matrix`''';
    __ducktype__ = (eggs.sp_coo_matrix, )

class csrarray(Duck):
    '''`scipy.sparse.csr_array`''';
    __ducktype__ = (eggs.sp_csr_array, )

class csrmatrix(Duck):
    '''`scipy.sparse.csr_matrix`''';
    __ducktype__ = (eggs.sp_csr_matrix, )

class diaarray(Duck):
    '''`scipy.sparse.dia_array`''';
    __ducktype__ = (eggs.sp_dia_array, )

class diamatrix(Duck):
    '''`scipy.sparse.dia_matrix`''';
    __ducktype__ = (eggs.sp_dia_matrix, )

class dokarray(Duck):
    '''`scipy.sparse.dok_array`''';
    __ducktype__ = (eggs.sp_dok_array, )

class dokmatrix(Duck):
    '''`scipy.sparse.dok_matrix`''';
    __ducktype__ = (eggs.sp_dok_matrix, )

class lilarray(Duck):
    '''`scipy.sparse.lil_array`''';
    __ducktype__ = (eggs.sp_lil_array, )

class lilmatrix(Duck):
    '''`scipy.sparse.lil_matrix`''';
    __ducktype__ = (eggs.sp_lil_matrix, )

class bsrarray(Duck):
    '''`scipy.sparse.bsr_array`''';
    __ducktype__ = (eggs.sp_bsr_array, )

class bsrmatrix(Duck):
    '''`scipy.sparse.bsr_matrix`''';
    __ducktype__ = (eggs.sp_bsr_matrix, )

# %% ../nbs/40_duck.ipynb 108
class sparsearray(Duck, Generic[*Ts]): 
    '''sparse array (e.g. 
    `scipy.sparse.sparray`, `scipy.sparse.cscarray`, `scipy.sparse.cooarray`, 
    `scipy.sparse.csrarray`, `scipy.sparse.diaarray`, `scipy.sparse.dokarray`, 
    `scipy.sparse.lilarray`, or `scipy.sparse.bsrarray`)'''
    __ducktype__ = (sparray, cscarray, cooarray, csrarray, diaarray, dokarray, lilarray, bsrarray)
    
    
class sparsematrix(Duck, Generic[*Ts]): 
    '''sparse type (e.g.  
    `scipy.sparse.spmatrix`, `scipy.sparse.cscmatrix`, `scipy.sparse.coomatrix`, 
    `scipy.sparse.csrmatrix`, `scipy.sparse.diamatrix`, `scipy.sparse.dokmatrix`, 
    `scipy.sparse.lilmatrix`, or `scipy.sparse.bsrmatrix`)'''
    __ducktype__ = (spmatrix, cscmatrix, coomatrix, csrmatrix, diamatrix, dokmatrix, lilmatrix, bsrmatrix)


# %% ../nbs/40_duck.ipynb 110
class sparse(Duck, Generic[*Ts]): 
    '''sparse type (e.g. `sparsearray` or `sparsematrix`)'''
    __ducktype__ = (sparray, sparsematrix)
    
class array(Duck, Generic[*Ts]): 
    '''array type (e.g. `np.array`, `np.recarray`, `sp.sparray`, `itertype`, or `list`)'''
    __ducktype__ = (nparray, recarray, sparray, itertype, list)
    
class matrix(Duck, Generic[*Ts]): 
    '''matrix type(e.g. `np.matrix`, `sp.matrix`, `itertype`, or `list`)'''
    __ducktype__ = (npmatrix, spmatrix, itertype, list)

# %% ../nbs/40_duck.ipynb 111
class sparseq(sparse[*Ts], Generic[*Ts]): 
    '''Optional sparse type (e.g. `sparsearray` or `sparsematrix`)'''
    __species__ = Species.OPT
    
class arrayq(array[*Ts], Generic[*Ts]): 
    '''Optional array type (e.g. `np.array`, `np.recarray`, `sp.sparray`, `itertype`, or `list`)'''
    __species__ = Species.OPT
    
class matrixq(matrix[*Ts], Generic[*Ts]): 
    '''Optional matrix type(e.g. `np.matrix`, `sp.matrix`, `itertype`, or `list`)'''
    __species__ = Species.OPT

# %% ../nbs/40_duck.ipynb 114
class cv2vidcap(Duck): 
    '''`cv2.VideoCapture`''';
    __ducktype__ = (eggs.cv_vidcap, )

class pilimage(Duck): 
    '''`PIL.Image.Image`''';
    __ducktype__ = (eggs.pil_image, )    

# %% ../nbs/40_duck.ipynb 116
class vidcap(Duck): 
    '''`cv2.VideoCapture`''';
    __ducktype__ = (eggs.cv_vidcap, )
    
class image(Duck): 
    '''`PIL.Image.Image`''';
    __ducktype__ = (eggs.pil_image, )

# %% ../nbs/40_duck.ipynb 119
class groupkey(Duck, Generic[*Ts]): 
    '''groupby key (e.g. `str`, `tuple`, or `numstr`)'''
    __ducktype__ = (str, tuple, numstr)
    

# %% ../nbs/40_duck.ipynb 121
class pddelegate(Duck):
    '''`pandas.core.accessor.PandasDelegate`''';
    __ducktype__ = (object, eggs.pd_delegate, )
    
class pdcataccessor(Duck):
    '''`pandas.core.arrays.categorical.CategoricalAccessor`''';
    __ducktype__ = (object, eggs.pd_cataccessor, )

# %% ../nbs/40_duck.ipynb 123
class pdindex(Duck, Generic[*Ts]):
    '''`pandas.Index`''';
    __ducktype__ = (eggs.pd_index, )
    
class pdcatindex(Duck, Generic[*Ts]):
    '''`pandas.CategoricalIndex`''';
    __ducktype__ = (eggs.pd_catindex, )

# %% ../nbs/40_duck.ipynb 126
class pdseries(Duck, Generic[*Ts]):
    '''`pandas.Series`''';
    __ducktype__ = (eggs.pd_series, )

# %% ../nbs/40_duck.ipynb 128
class boolindex(Duck, Generic[*Ts]):
    '''boolean index'''
    __ducktype__ = (list, pdindex, nparray, bools, )
    __expected__ = dict(dtype=bool01)
    
class catindex(Duck, Generic[*Ts]):
    '''categorical index'''
    __ducktype__ = (pdcatindex, )

# %% ../nbs/40_duck.ipynb 129
class boolindexq(boolindex[*Ts], Generic[*Ts]): 
    '''Optional[booleanindex]'''
    __species__ = Species.OPT
    
class catindexq(catindex[*Ts], Generic[*Ts]): 
    '''Optional[catindex]'''
    __species__ = Species.OPT

# %% ../nbs/40_duck.ipynb 131
class indextype(Duck): 
    '''index types(e.g. `pd.Index`, `pd.CategoricalIndex`, `np.ndarray`, or `boolindex`)'''
    __ducktype__ = (pdindex, pdcatindex, boolindex, nparray)
    
    
class indexlike(Duck): 
    '''index-like objects (e.g. `list`, `tuple`,`bools`, `nums`, 
    `itertype`, `pd.Index`, `pd.CategoricalIndex`, `boolindex`, or `np.array`)'''
    __ducktype__ = (list, tuple, bools, nums, itertype, pdindex, pdcatindex, boolindex, nparray)

# %% ../nbs/40_duck.ipynb 133
class seriestype(Duck, Generic[*Ts]):
    '''series type (e.g. `pd.Series`, `np.array`, `np.recarray`, `list`, or `indexlike`)'''
    __ducktype__ = (pdseries, nparray, recarray, list, indexlike)
    

class serieslike(Duck, Generic[*Ts]):
    '''series-like type'''
    __ducktype__ = (pdseries, nparray, recarray, list, indexlike)

# %% ../nbs/40_duck.ipynb 135
class pddataframe(Duck, Generic[*Ts]):
    '''`pandas.DataFrame`''';
    __ducktype__ = (eggs.pd_dataframe, )

# %% ../nbs/40_duck.ipynb 137
class dataframe(Duck, Generic[*Ts]):
    '''`pandas.DataFrame`''';
    __ducktype__ = (eggs.pd_dataframe, )

# %% ../nbs/40_duck.ipynb 140
class pttype(Duck, Generic[*Ts]):
    '''`torch.dtype`'''
    __ducktype__ = (eggs.pt_type, )
    
class pttypeq(pttype, Generic[*Ts]):
    '''`Optional[torch.dtype]`'''
    __species__ = Species.OPT

# %% ../nbs/40_duck.ipynb 142
class memfmt(Duck): 
    '''`torch.memory_format`''';
    __ducktype__ = (eggs.pt_memoryformat, )

class presfmt(Duck): 
    '''`torch.preserve_format`''';
    __ducktype__ = (eggs.pt_preserveformat, )

# %% ../nbs/40_duck.ipynb 144
class ptsize(Duck, Generic[*Ts]):
    '''`torch.Size`'''
    __ducktype__ = (eggs.pt_size, iterable, tuple, )
    
class ptsizeq(ptsize, Generic[*Ts]):
    '''`Optional[torch.Size]`'''
    __species__ = Species.OPT

# %% ../nbs/40_duck.ipynb 146
class ptdevice(Duck, Generic[*Ts]):
    '''`torch.device`'''
    __ducktype__ = (eggs.pt_device, iterable, tuple, )

class ptdeviceq(ptdevice, Generic[*Ts]):
    '''`Optional[torch.device]`'''
    __species__ = Species.OPT

# %% ../nbs/40_duck.ipynb 148
class pttensor(Duck, Generic[*Ts]): 
    '''`torch.Tensor`''';
    __ducktype__ = (eggs.pt_tensor, )
    
class inttensor(Duck, Generic[*Ts]): 
    '''`torch.IntTensor`''';
    __ducktype__ = (eggs.pt_inttensor, )
    
class booltensor(Duck, Generic[*Ts]): 
    '''`torch.BoolTensor`''';
    __ducktype__ = (eggs.pt_booltensor, )
    
class floattensor(Duck, Generic[*Ts]): 
    '''`torch.FloatTensor`''';
    __ducktype__ = (eggs.pt_floattensor, )
    
class longtensor(Duck, Generic[*Ts]): 
    '''`torch.LongTensor`''';
    __ducktype__ = (eggs.pt_longtensor, )
    
class doubletensor(Duck, Generic[*Ts]): 
    '''`torch.DoubleTensor`''';
    __ducktype__ = (eggs.pt_doubletensor, )

# %% ../nbs/40_duck.ipynb 149
class pttensorq(pttensor, Generic[*Ts]): 
    '''`Optional[torch.Tensor]`''';
    __species__ = Species.OPT
    
class inttensorq(inttensor, Generic[*Ts]): 
    '''`Optional[torch.IntTensor]`''';
    __species__ = Species.OPT
    
class booltensorq(booltensor, Generic[*Ts]): 
    '''`Optional[torch.BoolTensor]`''';
    __species__ = Species.OPT
    
class floattensorq(floattensor, Generic[*Ts]): 
    '''`Optional[torch.FloatTensor]`''';
    __species__ = Species.OPT
    
class longtensorq(longtensor, Generic[*Ts]): 
    '''`Optional[torch.LongTensor]`''';
    __species__ = Species.OPT
    
class doubletensorq(doubletensor, Generic[*Ts]): 
    '''`Optional[torch.DoubleTensor]`''';
    __species__ = Species.OPT

# %% ../nbs/40_duck.ipynb 151
class nantensor(Duck, Generic[*Ts]): 
    '''tensor of `nan`s'''
    __ducktype__ = (pttensor, nans, )
    __expected__ = dict(dtype=nan)
    

class nantensorq(nantensor[*Ts], Generic[*Ts]): 
    '''Optional tensor of `nan`s'''
    __species__ = Species.OPT

# %% ../nbs/40_duck.ipynb 153
class tensor(Duck, Generic[*Ts]): 
    '''tensor type (e.g. `torch.tensor`, `torch.IntTensor`, `torch.BoolTensor`, or `torch.FloatTensor`)'''
    __ducktype__ = (pttensor, inttensor, booltensor, floattensor, longtensor, doubletensor, )
    
class tensorq(tensor, Generic[*Ts]): 
    '''`Optional[tensor]`'''
    __species__ = Species.OPT
    
class pttensorq(pttensor, Generic[*Ts]): 
    '''`Optional[pttensor]`'''
    __species__ = Species.OPT

# %% ../nbs/40_duck.ipynb 155
class ptmodule(Duck): 
    '''`torch.nn.Module`'''
    __ducktype__ = (eggs.pt_module, )

class plmodule(Duck): 
    '''`pytorch_lightning.LightningModule`'''
    __ducktype__ = (eggs.pl_module, )

# %% ../nbs/40_duck.ipynb 157
class nnmodule(Duck): 
    '''module type (e.g. `torch.nn.Module` or `pytorch_lightning.LightningModule`)'''
    __ducktype__ = (ptmodule, plmodule, )

# %% ../nbs/40_duck.ipynb 159
class ptdataset(Duck): 
    '''`torch.utils.data.Dataset`'''
    __ducktype__ = (eggs.pt_dataset, )

class pldatamodule(Duck): 
    '''`pytorch_lightning.LightningDataModule`'''
    __ducktype__ = (eggs.pl_datamodule, )

class ptdataloader(Duck): 
    '''`torch.utils.data.DataLoader`'''
    __ducktype__ = (eggs.pt_dataloader, )


# %% ../nbs/40_duck.ipynb 162
class neuralode(Duck): 
    '''torchdyn neural ODE'''
    __ducktype__ = (eggs.td_neuralode, )

class neuralsde(Duck): 
    '''torchdyn neural SDE'''
    __ducktype__ = (eggs.td_neuralsde, )

class neuraldynamics(Duck): 
    '''neuralode or neuralsde'''
    __ducktype__ = (neuralode, neuralsde, )


# %% ../nbs/40_duck.ipynb 163
class augmenter(Duck): 
    '''`torchdyn.nn.augmenter`''';
    __ducktype__ = (eggs.td_augmenter, )

# %% ../nbs/40_duck.ipynb 164
class odefunc(Duck): 
    '''`torchdyn.ODEFunc`'''
    __ducktype__ = (eggs.td_odefunc, )

class sdefunc(Duck): 
    '''`torchdyn.SDEFunc`'''
    __ducktype__ = (eggs.td_sdefunc, )

# %% ../nbs/40_duck.ipynb 166
class sdeito(Duck): 
    '''`torchsde.ito`'''
    __ducktype__ = (eggs.ts_sdeito, )

class brownianintervial(Duck): 
    '''`torchsde.BrownianInterval`'''
    __ducktype__ = (eggs.ts_brownianinterval, )


# %% ../nbs/40_duck.ipynb 169
class dtype(Duck, Generic[*Ts]):
    '''Type of data (e.g. `np.dtype` or `torch.dtype`)'''
    __ducktype__ = (type, eggs.np_type, eggs.pt_type, )
    
class dtypeq(dtype[*Ts]):
    '''Optional type of data (e.g. `np.dtype` or `torch.dtype`)'''
    __species__ = Species.OPT

# %% ../nbs/40_duck.ipynb 171
class device(Duck, Generic[*Ts]):
    '''`torch.device`'''
    __ducktype__ = (eggs.pt_device, str, )
    __dev__: ClassVar[str] = 'cpu'
    
    @classmethod
    def __metahook__(cls, mcls, name, bases, namespace, **kwargs):
        super().__metahook__(mcls, name, bases, namespace, **kwargs)
        d = cls.__dev__
        isdev = grds._isdev(d)
        cls.__expected__.update(__dev__=d)
        cls.__extrafns__.update(__dev__=isdev)
        return cls
    
class deviceq(device, Generic[*Ts]): 
    '''`Optional [torch.device]`'''
    __species__ = Species.OPT
    @classmethod
    def __generate__(cls):
        dev = getattr(cls, __DEV__, None)
        try: return eggs.pt_device.duck(dev)
        except: return dev or None

# %% ../nbs/40_duck.ipynb 175
class devcpu(device[Literal['cpu']]):
    '''`torch.device['cpu']`'''
    __dev__: ClassVar[str] = 'cpu'
    
class devmps(device[Literal['mps']]):
    '''`torch.device['mps']`'''
    __dev__: ClassVar[str] = 'mps'
    
class devcuda(device[Literal['cuda']]):
    '''`torch.device['mps']`'''
    __dev__: ClassVar[str] = 'cuda'

# %% ../nbs/40_duck.ipynb 177
class size(itertype[*Ts], Generic[*Ts]): 
    '''size type (e.g. `tuple`, `torch.Size`, `itertype`)'''
    __ducktype__ = (tuple, ptsize, itertype, )
    __expected__ = dict(dtype=int)
    
    
class sizeq(size[*Ts]): 
    '''Optional size type (e.g. `tuple`, `torch.Size`, `itertype`)'''
    __species__ = Species.OPT

# %% ../nbs/40_duck.ipynb 179
class datasource(Duck, Generic[*Ts]): 
    '''`numpy.DataSource`''';
    __ducktype__ = (eggs.np_datasource, )   

# %% ../nbs/40_duck.ipynb 181
class adata(Duck): 
    '''`anndata.AnnData`''';
    __ducktype__ = (eggs.ad_adata, )

class adatas(Duck, Generic[*Ts]): 
    '''`anndata.AnnData`''';
    __ducktype__ = (itertype, )
    __expected__ = dict(dtype=adata)

# %% ../nbs/40_duck.ipynb 183
class graphtype(Duck, Generic[*Ts]): 
    '''graph type (e.g. `BaseGraph`, `DataGraph`, `LandmarkGraph`, `kNNGraph`, `TraditionalGraph`, 
    `MNNGraph`, `kNNLandmarkGraph`, `TraditionalLandmarkGraph`, or `MNNLandmarkGraph`)'''
    __ducktype__ = (
        eggs.gt_bgraph, eggs.gt_dgraph, eggs.gt_lgraph, 
        eggs.gt_kgraph, eggs.gt_tgraph, eggs.gt_mgraph, 
        eggs.gt_klgraph, eggs.gt_tlgraph, eggs.gt_mlgraph
    )

# %% ../nbs/40_duck.ipynb 185
class datatype(Duck, Generic[*Ts]):
    '''data type (e.g.  `np.array`, `np.matrix`, `sp.array`, `sp.matrix`,  `np.recarray`, or `anndata`)'''
    __ducktype__ = (array, matrix, adata, )


# %% ../nbs/40_duck.ipynb 187
class nnum(nums, nlen[*Ts]):
    @classmethod
    def __metahook__(cls, mcls, name, bases, namespace, **kwargs):
        super().__metahook__(mcls, name, bases, namespace, **kwargs)
        cls.__expected__.update(dtype=num)
        return cls

class nint(ints, nlen[*Ts]):
    @classmethod
    def __metahook__(cls, mcls, name, bases, namespace, **kwargs):
        super().__metahook__(mcls, name, bases, namespace, **kwargs)
        cls.__expected__.update(dtype=int)
        return cls
    
class nfloat(floats, nlen[*Ts]):
    @classmethod
    def __metahook__(cls, mcls, name, bases, namespace, **kwargs):
        super().__metahook__(mcls, name, bases, namespace, **kwargs)
        cls.__expected__.update(dtype=float)
        return cls

# %% ../nbs/40_duck.ipynb 189
class int2(nint[int, int], len2): 
    '''A ducktype for an iterable of ints with length 2'''

class int3(nint[int, int, int], len3): 
    '''A ducktype for an iterable of ints with length 3'''


class int4(nint[int, int, int, int], len4): 
    '''A ducktype for an iterable of ints with length 4'''


# %% ../nbs/40_duck.ipynb 193
class shaped(Duck, Generic[*Ts]): 
    '''shaped `datatype`'''
    __ducktype__ = (datatype, indextype, seriestype, dataframe, tensor, )
    __required__ = (SHAPE, )

# %% ../nbs/40_duck.ipynb 195
class dimmed(Duck, Generic[*Ts]):
    '''`np.ndarray` with shape `(i, ...)`'''
    __ducktype__ = (list, nparray, list, itertype, )
    __required__ = (NDIM, )

class dim1(dimmed, Generic[*Ts]):
    '''`np.ndarray` with shape `(i, )`'''
    __expected__ = dict(ndim=1)
    

class dim2(dimmed, Generic[*Ts]):
    '''`np.ndarray` with shape `(i, j, )`'''
    __expected__ = dict(ndim=2)
    

class dim3(dimmed, Generic[*Ts]):
    '''`np.ndarray` with shape `(i, j, k, )`'''
    __expected__ = dict(ndim=3)

class dim4(dimmed, Generic[*Ts]):
    '''`np.ndarray` with shape `(i, j, k, l, )`'''
    __expected__ = dict(ndim=4)

# %% ../nbs/40_duck.ipynb 197
class genset(Duck[set], Generic[*Ts]):
    '''A type alias for a generic set type.''';
    __ducktype__ = (set, )
    
class setstr(genset[str]):
    '''A type alias for a set of strings.''';
    __expected__ = dict(dtype=str)
    
class setnum(genset[num]):
    '''A type alias for a set of numbers.''';
    __expected__ = dict(dtype=num)

# %% ../nbs/40_duck.ipynb 199
class kvdict(Duck[dict[T, O]], Generic[T, O]):
    '''A type alias for a generic set type.''';
    __ducktype__ = (dict, )
    __required__ = (KEYS, VALUES, )
    
    __k__: ClassVar[Type[T]] = str
    __v__: ClassVar[Type[T]] = notnone
    __kvtypes__: ClassVar[tuple[T, O]] = (str, notnone)
    
    @classmethod
    def __metahook__(cls, mcls, name, bases, namespace, **kwargs):
        super().__metahook__(mcls, name, bases, namespace, **kwargs)
        
        t = getattr(cls, __KVTYPES__, (None, None))
        k = getattr(cls, __K__, None)
        v = getattr(cls, __V__, None)
        if t == (None, None): t = (k, v)
        if k is None: k = t[0]
        if v is None: v = t[1]
        
        if not hasattr(cls, __KVTYPES__): setattr(cls, __KVTYPES__, t)
        if not hasattr(cls, __K__): setattr(cls, __K__, k)
        if not hasattr(cls, __V__): setattr(cls, __V__, v)
        istmap = grds._iskvdict(cls)
        cls.__expected__.update(__k__=k, __v__=v, __kvtypes__=t)
        cls.__extrafns__.update(__k__=istmap, __v__=istmap, __kvtypes__=istmap)
        return cls
    

# %% ../nbs/40_duck.ipynb 201
class pttndict(kvdict[str, pattern]): 
    '''A type alias for a dictionary of named regex patterns dict[str, re.Pattern]''';
    __ducktype__ = (dict, )
    __k__: ClassVar[Type[T]] = str
    __v__: ClassVar[Type[T]] = pattern
    __kvtypes__: ClassVar[tuple[T, O]] = (str, pattern)

# %% ../nbs/40_duck.ipynb 203
class annreturn(Duck[Callable[P, T]]): 
    '''A type alias for an Annotated Return of a Callable.'''
    __ducktype__ = (Callable, )
    __retannot__: ClassVar[T] = bool01
    
    @classmethod
    def __guard__(cls, obj: Any) -> bool:
        return super().__instancecheck__(obj) and getretann(obj) in cls.__ducktype__
    
    
    @classmethod
    def __metahook__(cls, mcls, name, bases, namespace, **kwargs):
        super().__metahook__(mcls, name, bases, namespace, **kwargs)
        
        t = getattr(cls, __RETANNOT__, bool)
        if not hasattr(cls, __RETANNOT__): 
            setattr(cls, __RETANNOT__, t)
            
        def isretann(obj, val) -> TypeGuard[T]:
            return getretann(obj) in (cls.__retannot__, *cls.__ducktype__, )
        
        cls.__expected__.update(__retannot__=t)
        cls.__extrafns__.update(__retannot__=isretann)
        return cls

# %% ../nbs/40_duck.ipynb 205
class slicelike(Duck):
    '''A type alias for a slice or the integers corresponding to the
    `(start, stop, step)` of a slice i.e. `slice | tuple[intq, int, intq]`
    
    Notes
    -----
    slice : SliceFormats
        The default representations of a python slice, i.e. 
        `(start: int, stop: int, step: int)`
        
    coord : SliceFormats
        The coordinate repesentation of a python slice i.e. 
        `(x0, y0, ...), (x1, y1, ...), ...`
        
    __nones_ok : bool
        A sometimes optional parameter for slice utility functions.
        If `True`, the result is not corrected to prevent None values.
        i.e. `slice(0, 10, None)` would become `slice(0, 10, 1)`
        
                
    Definitions
    -----------
    :slice: `slice`
        The standard representation for a python slice, e.g.
        `(start: int, stop: int, step: int)` or `slice(0, 10, 2)`
    
    
    :svals: `tuple[int | None, int, int | None]`
        The integer representation for a python slice, e.g.
        `(start: int | None, stop: int, step: int | None)` or `slice(0, 10, None)`
        
    :slices: `tuple[slice | svals, ...] `
        An iterable of like representation for a python slice, e.g.
        `slice(0, 10, 2)`, `tuple(0, 10, 2)`, `tuple(0, 10, None)`, etc
    
    :coord: `tuple[int, ...]`
        The coordinate representation for a python slice, e.g.
        `(x0, y0, ...), (x1, y1, ...), ...`
        
    :coords: `tuple[coord, ...]`
        An iterable of like coordinate representation for a python slice, e.g.
        `(x0, y0, ...), (x1, y1, ...), ...`
    
    Examples
    -------- 
    >>> slices = (0, 5), (2, 10) # slice format: (x0, x1), (y0, y1)
    ... tocoordformat(slices) # convert to coordinate format
    ((0, 2), (5, 10)) 

    >>> coords = (0, 2), (5, 10) # coord format: (x0, y0), (x1, y1)
    ... tosliceformat(coords) # convert to slice format
    ((0, 5), (2, 10))
    
    See Also
    --------
    asto.slcs.tocoordformat: convert slices in slice format to coordinate format
    asto.slcs.tosliceformat: convert slices in coordinate format to slice format
    '''
    __ducktype__ = (slice, int3)
    
    
class slicespec(Duck):
    '''A type alias for a slice or one of two formats to specify a slice.
    
    Notes
    -----
    slice : SliceFormats
        The default representations of a python slice, i.e. 
        `(start: int, stop: int, step: int)`
        
    coord : SliceFormats
        The coordinate repesentation of a python slice i.e. 
        `(x0, y0, ...), (x1, y1, ...), ...`
        
    __nones_ok : bool
        A sometimes optional parameter for slice utility functions.
        If `True`, the result is not corrected to prevent None values.
        i.e. `slice(0, 10, None)` would become `slice(0, 10, 1)`
        
                
    Definitions
    -----------
    :slice: `slice`
        The standard representation for a python slice, e.g.
        `(start: int, stop: int, step: int)` or `slice(0, 10, 2)`
    
    
    :svals: `tuple[int | None, int, int | None]`
        The integer representation for a python slice, e.g.
        `(start: int | None, stop: int, step: int | None)` or `slice(0, 10, None)`
        
    :slices: `tuple[slice | svals, ...] `
        An iterable of like representation for a python slice, e.g.
        `slice(0, 10, 2)`, `tuple(0, 10, 2)`, `tuple(0, 10, None)`, etc
    
    :coord: `tuple[int, ...]`
        The coordinate representation for a python slice, e.g.
        `(x0, y0, ...), (x1, y1, ...), ...`
        
    :coords: `tuple[coord, ...]`
        An iterable of like coordinate representation for a python slice, e.g.
        `(x0, y0, ...), (x1, y1, ...), ...`
    
    Examples
    -------- 
    >>> slices = (0, 5), (2, 10) # slice format: (x0, x1), (y0, y1)
    ... tocoordformat(slices) # convert to coordinate format
    ((0, 2), (5, 10)) 

    >>> coords = (0, 2), (5, 10) # coord format: (x0, y0), (x1, y1)
    ... tosliceformat(coords) # convert to slice format
    ((0, 5), (2, 10))
    
    See Also
    --------
    asto.slcs.tocoordformat: convert slices in slice format to coordinate format
    asto.slcs.tosliceformat: convert slices in coordinate format to slice format
    '''
    __ducktype__ = (slicelike, tuple, tuple[int, ...])
    
    
class slicespecs(itertype[slicespec]):
    '''A type alias for a list of slices, the integers corresponding to the
    `(start, stop, step)` of a slice or a slicespec.
    
    Notes
    -----
    slice : SliceFormats
        The default representations of a python slice, i.e. 
        `(start: int, stop: int, step: int)`
        
    coord : SliceFormats
        The coordinate repesentation of a python slice i.e. 
        `(x0, y0, ...), (x1, y1, ...), ...`
        
    __nones_ok : bool
        A sometimes optional parameter for slice utility functions.
        If `True`, the result is not corrected to prevent None values.
        i.e. `slice(0, 10, None)` would become `slice(0, 10, 1)`
        
                
    Definitions
    -----------
    :slice: `slice`
        The standard representation for a python slice, e.g.
        `(start: int, stop: int, step: int)` or `slice(0, 10, 2)`
    
    
    :svals: `tuple[int | None, int, int | None]`
        The integer representation for a python slice, e.g.
        `(start: int | None, stop: int, step: int | None)` or `slice(0, 10, None)`
        
    :slices: `tuple[slice | svals, ...] `
        An iterable of like representation for a python slice, e.g.
        `slice(0, 10, 2)`, `tuple(0, 10, 2)`, `tuple(0, 10, None)`, etc
    
    :coord: `tuple[int, ...]`
        The coordinate representation for a python slice, e.g.
        `(x0, y0, ...), (x1, y1, ...), ...`
        
    :coords: `tuple[coord, ...]`
        An iterable of like coordinate representation for a python slice, e.g.
        `(x0, y0, ...), (x1, y1, ...), ...`
    
    Examples
    -------- 
    >>> slices = (0, 5), (2, 10) # slice format: (x0, x1), (y0, y1)
    ... tocoordformat(slices) # convert to coordinate format
    ((0, 2), (5, 10)) 

    >>> coords = (0, 2), (5, 10) # coord format: (x0, y0), (x1, y1)
    ... tosliceformat(coords) # convert to slice format
    ((0, 5), (2, 10))
    
    See Also
    --------
    asto.slcs.tocoordformat: convert slices in slice format to coordinate format
    asto.slcs.tosliceformat: convert slices in coordinate format to slice format
    ''';
    __ducktype__ = (slicespec, )

# %% ../nbs/40_duck.ipynb 208
class richtree(Duck):
    '''`rich.tree.Tree`'''
    __ducktype__ = (eggs.r_richtree, )
    
class richtext(Duck):
    '''`rich.text.Text`'''
    __ducktype__ = (eggs.r_richtext, )
    
class richcons(Duck):
    '''`rich.console.Console`'''
    __ducktype__ = (eggs.r_richcons, )
    
class richprog(Duck):
    '''`rich.progress.Progress`'''
    __ducktype__ = (eggs.r_richprog, )
    
class richtask(Duck):
    '''`rich.progress.Task`'''
    __ducktype__ = (eggs.r_richtask, )

# %% ../nbs/40_duck.ipynb 210
class tqdmauto(Duck):
    '''`tqdm.auto.tqdm`'''
    __ducktype__ = (eggs.tq_as, )

# %% ../nbs/40_duck.ipynb 212
class typerapp(Duck):
    '''`typer.Typer`'''
    __ducktype__ = (eggs.t_typer, )
    



# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/06_solo.ipynb.

# %% auto 0
__all__ = ['Singleton', 'Empty', 'Module', 'EmptyModule']

# %% ../nbs/06_solo.ipynb 6
#| export


# %% ../nbs/06_solo.ipynb 8
from types import ModuleType
from typing import (Any, )

# %% ../nbs/06_solo.ipynb 10
#| export

# %% ../nbs/06_solo.ipynb 12
#| export

# %% ../nbs/06_solo.ipynb 14
#| export


# %% ../nbs/06_solo.ipynb 16
class Singleton(type):
    _ins = {}
    def __new__(cls, name: str = 'Singleton', bases: tuple = (), namespace: dict = {}, **kwargs):
        insdict = namespace.get('_ins', cls._ins)
        insdict.update(cls._ins)
        namespace.update(_ins=insdict)
        if name not in insdict:
            cls._ins[name] = super().__new__(cls, name, bases, namespace, **kwargs)
        return cls._ins[name]
        
    def __instancecheck__(self, __instance: Any) -> bool: return super().__instancecheck__(__instance)    
    def __subclasscheck__(self, __subclass: type) -> bool: return super().__subclasscheck__(__subclass)
    def __eq__(cls, ins) -> bool:  return isinstance(ins, (cls, cls.__class__))
    def __hash__(cls) -> int: return hash(cls.__name__)
    def __repr__(cls) -> str: return f'<{type(cls).__name__.lower()} {cls.__name__}>'
    def __call__(cls, *args, **kwargs): return cls._ins[cls.__name__]

    
class Empty(Singleton): ...
class Module(ModuleType, metaclass=Empty): ...
class EmptyModule(ModuleType, metaclass=Singleton): ... 



__version__ = "0.0.38"
# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/00__init__.ipynb.

# %% auto 0
__all__ = ['loadmod']

# %% ../nbs/00__init__.ipynb 3
from itertools import chain
from types import ModuleType
from importlib import import_module

def loadmod(mod: ModuleType):
    if hasattr(mod, '__all__'): return mod.__all__
    if hasattr(mod, '__module__'): return loadmod(import_module(mod.__module__))
    raise AttributeError(f"module {mod} has no attribute '__all__'")

# %% ../nbs/00__init__.ipynb 4
from . import (cons, type, prot, grds, enum, errs, solo, misc, spec, meta, base, mods, eggs, duck, chic)
from .cons import *
from .type import *
from .prot import *
from .enum import *
from .errs import *
from .solo import *
from .grds import *
from .misc import *
from .spec import *
# from quac.meta import *
from .base import *
# from quac.mods import *
# from quac.eggs import *
from .duck import *
from .chic import *

# %% ../nbs/00__init__.ipynb 5
__all__ = sorted(set(chain(*[
    loadmod(mod) for mod in (cons, type, prot, grds, enum, errs, solo, misc, spec, base, duck, chic)
])))



# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/50_chic.ipynb.

# %% auto 0
__all__ = ['boolfunc', 'flagfunc', 'numeric', 'numbers', 'emptystring', 'notemptystring', 'pttn', 'dirpath', 'dirnames',
           'filenames', 'enummember', 'layer', 'boundingbox', 'coordinate', 'coord', 'point', 'xypos', 'width',
           'height', 'rect', 'clsint', 'clsset', 'clsstr', 'clsobj', 'clsbool', 'clsbool01', 'clscall', 'clstype',
           'clslist', 'clsfloat', 'clsnum', 'clscomplex', 'slc', 'slcs', 'path', 'boolidx', 'bidx', 'catidx', 'cidx',
           'anndata', 'anndatas', 'series', 'index', 'lit', 'boolcall', 'boolcallq', 'boolcalls', 'docgenfn',
           'axesidxs', 'axesidxsq', 'argstype', 'sysexcinfotype', 'sysformatstyle', 'bbox', 'rgb', 'rgba', 'color',
           'imagesize', 'videosize', 'imagesizeq', 'videosizeq', 'mixin']

# %% ../nbs/50_chic.ipynb 6
from enum import member, _EnumDict

# %% ../nbs/50_chic.ipynb 8
from types import ModuleType, TracebackType

from typing import (
    Any, Dict, List, Set, Self, Type, Union, Tuple, Generic, Callable, Literal, 
    TypeVar, ClassVar, TypeAlias, ParamSpec, Optional,
    Iterable, ForwardRef, Mapping, TypeGuard
)

# %% ../nbs/50_chic.ipynb 10
#| export

# %% ../nbs/50_chic.ipynb 12
#| export

# %% ../nbs/50_chic.ipynb 14
from .cons import RETURN
from .type import P, T, Ts, O
from .enum import Species
from .misc import getretann
from .base import Duck

from . import grds, duck

# %% ../nbs/50_chic.ipynb 17
# bool01 = TypeVar('bool01', bound=duck.bool01)

# %% ../nbs/50_chic.ipynb 19
class boolcall(duck.annreturn[P, duck.bool01]):
    '''A type alias for a boolean or a callable that returns a boolean.'''
    __ducktype__ = (duck.bool01, Callable)
    @classmethod
    def __guard__(cls, obj: Any) -> bool:
        return super().__instancecheck__(obj) and getretann(obj) in (bool, duck.bool01)
  
class boolcallq(boolcall): 
    '''A type alias for a boolean or a callable that returns a boolean.'''
    __species__ = Species.OPT
      
class boolcalls(duck.bools, Duck[List[Callable[P, duck.bool01]]]): 
    '''A type alias for a list of booleans or callables that each returns a boolean.''';
    __expected__ = dict(dtype=boolcall)
    
boolfunc: TypeAlias = boolcall
'''A type alias for a function that returns a boolean.''';

flagfunc: TypeAlias = boolfunc
'''A type alias for a "flag" function i.e. a function that returns a boolean.''';

# %% ../nbs/50_chic.ipynb 20
numeric: TypeAlias = duck.num
'''A type alias for a "numeric" value i.e. `int`, `float`, `complex`, or `Number`''';

numbers: TypeAlias = duck.nums
'''A type alias for a list of "numeric" values i.e. `int`, `float`, `complex`, or `Number`''';

# %% ../nbs/50_chic.ipynb 22
emptystring = duck.nilstr
'''A type alias for the empty string `''` (aka the nil string)''';

notemptystring = duck.notnilstr
'''A type alias for not emptystring (not `''`, the empty string)''';

pttn: TypeAlias = duck.pattern
'''A type alias for `re.Pattern`''';

# %% ../nbs/50_chic.ipynb 23
class docgenfn(duck.annreturn[P, str]): 
    '''A type alias for a callable that returns as string, which should be a valid docstring.''';
    __ducktype__ = (Callable, )
    __retannot__ = str

# %% ../nbs/50_chic.ipynb 25
dirpath: TypeAlias = duck.pathlike
'''A type alias for a path to a directory.''';

dirnames: TypeAlias = duck.paths
'''A type alias for a list of paths to directories.''';

filenames: TypeAlias = duck.paths
'''A type alias for a list of paths to filenames.''';

# %% ../nbs/50_chic.ipynb 27
enummember: TypeAlias = duck.enumauto
'''A type alias for an `auto()` generated enum `member`.''';

# %% ../nbs/50_chic.ipynb 29
class axesidxs(duck.ints, Generic[*Ts]): 
    '''A type alias for an integer or a tuple of integers representing 
    the index / indicies of the axes of a tensor.'''
    __ducktype__ = (int, duck.ints, )
    
class axesidxsq(duck.ints, Generic[*Ts]): 
    '''A type alias for an integer or a tuple of integers representing 
    the index / indicies of the axes of a tensor.'''
    __species__ = Species.OPT
    

layer: TypeAlias = duck.ptmodule
'''A type alias for a PyTorch layer.''';

# %% ../nbs/50_chic.ipynb 31
class argstype(Duck[type[tuple[object, ...]] | type[Mapping[str, object]]]):
    '''A type alias for argument types.''';
    __ducktype__ = (tuple, Mapping[str, object], Mapping, )
    __expected__ = dict(dtype=object)
    
    
class sysexcinfotype(Duck[
    type[tuple[type[BaseException], BaseException, TracebackType | None]] | \
    type[tuple[None, None, None]]
]):
    '''A type alias for `SysExcInfoType` types i.e.
    `(type[BaseException], BaseException, TracebackType | None)` or `(None, None, None)`''';
    __ducktype__ = (type(BaseException), BaseException, Optional[TracebackType], )
    __expected__ = dict(dtype=object)
    
    
class sysformatstyle(Duck[Literal['%', '{', '$']]):
    __ducktype__ = ('%', '{', '$', )

# %% ../nbs/50_chic.ipynb 33
class bbox(duck.itertype[T: duck.int2, T: duck.int2], Generic[T, O]):
    '''A type alias representing a BoundingBox i.e. a tuple of 
    `(upper_left: int, upper_right: int), (lower_left: int, lower_right: int)`.'''; 
    __expected__ = dict(dtype=duck.int2)
   
    
boundingbox: TypeAlias = bbox[duck.int2, duck.int2]
 
coordinate: TypeAlias = duck.nums
'''A type alias for a coordinate i.e. a tuple of numbers representing a point in space.''';

coord: TypeAlias = coordinate
'''A type alias for a coordinate (`coord`) i.e. a tuple of numbers representing a point in space.''';

point: TypeAlias = coordinate
'''A type alias for a point i.e. $(x_1, x_2, ..., x_n)$.''';

xypos: TypeAlias = duck.int2
'''A type alias for an 2D-point in space i.e. `(x, y)`.'''; 

width: TypeAlias = int
'''the width of an object''';

height: TypeAlias = int
'''the height of an object''';

rect: TypeAlias = duck.int2
'''A type alias an 2D tuple representing the `width` and `height` of rectangle.'''; 

# %% ../nbs/50_chic.ipynb 35
class rgb(duck.int3, Generic[*Ts]):
    '''A type alias for an RGB color tuple `(r: int, g: int, b: int)` 
    where, `r`, `g`, & `b` are each integers between 0 and 255.'''
    @classmethod
    def __guard__(cls, obj: Any) -> bool:
        return super().__guard__(obj) and all(0 <= obj[i] <= 255 for i in range(3))
        
class rgba(duck.len4, duck.size[int, int, int, float], Generic[*Ts]):
    '''A type alias for a RGBA color tuple `(r: int, g: int, b: int, a: float)` 
    where, `r`, `g`, & `b` are each integers between 0 and 255 and `a` is a float between 0 and 1.0.'''
    @classmethod
    def __guard__(cls, obj: Any) -> bool:
        return duck.size(obj) and len(obj) == 4, rgb(obj[:-1]) and 0 <= obj[-1] <= 1
    
class color(Duck[rgb | rgba], Generic[*Ts]):
    '''A type alias for either a RGB or a RGBA color tuple `(r: int, g: int, b: int, a: float)` 
    where, `r`, `g`, & `b` are each integers between 0 and 255 and `a`
    is a float between 0 and 1.0.'''
    __ducktype__ = (rgb, rgba,)

# %% ../nbs/50_chic.ipynb 37
class imagesize(Duck[duck.int2, duck.int3]):
    '''A type alias for an image size i.e. a tuple of `(width: int, height: int)` or
    `(width: int, height: int, channels: int)`.'''
    __ducktype__ = (duck.int2, duck.int3, )

class videosize(Duck[duck.int3, duck.int4]):
    '''A type alias for a video size i.e. a tuple of `(width: int, height: int, channels: int)` or
    `(width: int, height: int, channels: int, frames: int)`.'''
    __ducktype__ = (duck.int3, duck.int4, )

class imagesizeq(imagesize):
    '''A type alias for an optional image size i.e. a tuple of `(width: int, height: int)` or
    `(width: int, height: int, channels: int)`.'''
    __species__ = Species.OPT
    
class videosizeq(videosize):
    '''A type alias for an optional video size i.e. a tuple of `(width: int, height: int, channels: int)` or
    `(width: int, height: int, channels: int, frames: int)`.'''
    __species__ = Species.OPT

# %% ../nbs/50_chic.ipynb 39
class mixin(Duck[Union[Tuple[Type, Callable], 'MixinType']]):
    __ducktype__ = (tuple, 'MixinType', )
    @classmethod
    def __guard__(cls, obj: Any) -> bool:
        if super().__instancecheck__(obj): return True
        return grds.istuple(obj) and len(obj) == 2 and isinstance(obj[0], type) and callable(obj[1])

# %% ../nbs/50_chic.ipynb 41
clsint = ClassVar[int]
'''A type alias for a class variable of type `int`.''';

clsset = ClassVar[set]
'''A type alias for a class variable of type `set`.''';

clsstr = ClassVar[str]
'''A type alias for a class variable of type `str`.''';

clsobj = ClassVar[object]
'''A type alias for a class variable of type `object`.''';

clsbool = ClassVar[bool]
'''A type alias for a class variable of type `bool`.''';

clsbool01 = ClassVar[duck.bool01]
'''A type alias for a class variable of type `bool01`.''';

clscall = ClassVar[Callable]
'''A type alias for a class variable of type `Callable`.''';

clstype = ClassVar[type]
'''A type alias for a class variable of type `type`.''';

clslist = ClassVar[list]
'''A type alias for a class variable of type `list`.''';

clsfloat = ClassVar[float]
'''A type alias for a class variable of type `float`.''';

clsnum = ClassVar[duck.num]
'''A type alias for a class variable of type `num`.''';

clscomplex = ClassVar[complex]
'''A type alias for a class variable of type `complex`.''';


# %% ../nbs/50_chic.ipynb 43
slc: TypeAlias = duck.slicespec
'''A type alias for a slice or the integers corresponding to the
`(start, stop, step)` of a slice.'''
    
slcs: TypeAlias = duck.slicespecs
'''A type alias for a list of slices or the integers corresponding to the
`(start, stop, step)` of a slice.''';

# %% ../nbs/50_chic.ipynb 45
path: TypeAlias = duck.pathlike
'''A type alias for `os.PathLike`, `pathlib.Path`, `pathlib.PosixPath`''';

# %% ../nbs/50_chic.ipynb 47
boolidx: TypeAlias = duck.boolindex
'''A type alias for a boolean index.''';
bidx: TypeAlias = duck.boolindex
'''A type alias for a boolean index.''';
catidx: TypeAlias = duck.catindex
'''A type alias for a categorical index.''';
cidx: TypeAlias = duck.catindex
'''A type alias for a categorical index.''';


# %% ../nbs/50_chic.ipynb 49
anndata: TypeAlias = duck.adata
'''A type alias for an `AnnData` object.''';

anndatas: TypeAlias = duck.adatas
'''A type alias for a list of  `AnnData` object.''';

series: TypeAlias = duck.pdseries
'''A type alias for a `pandas.Series` object.''';

index: TypeAlias = duck.pdindex
'''A type alias for a `pandas.Index` object.''';

# %% ../nbs/50_chic.ipynb 51
lit: TypeAlias = duck.literal
'''A type alias for a `Literal` type.''';



# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/09_meta.ipynb.

# %% auto 0
__all__ = ['ABCSpecies', 'ABCDuck', 'DuckProtocolMeta', 'DuckModuleMeta']

# %% ../nbs/09_meta.ipynb 6
import sys, copy, itertools
from abc import ABCMeta, abstractmethod
from functools import partial
from inspect import isclass
from importlib import import_module
from importlib.util import (module_from_spec)
from importlib.machinery import ModuleSpec

# %% ../nbs/09_meta.ipynb 8
from types import ModuleType, NoneType, FunctionType
from typing import (Any, Self, Callable, TypeGuard, ClassVar, _ProtocolMeta)

# %% ../nbs/09_meta.ipynb 10
#| export

# %% ../nbs/09_meta.ipynb 12
#| export

# %% ../nbs/09_meta.ipynb 14
from quac.cons import (
    ANNOYING_ATTRS, MODULE_TYPE_ATTRS,
    _DUCK, ALL_ATTRS,
    __BASES__, __PROPS__, __MODULE__, __QUALNAME__, __ANNOTATIONS__,  
    __HUSH__, __DUCK__, __ALTS__, __DUCKMODULE__, __DUCKTYPES__,
    __GENERATE__, __DUCKTYPE__,
    CALL_INIT_ATTRS
)
from .type import T, O, P, Ts
from .solo import EmptyModule
from .prot import NotProtocol, OptProtocol
from .enum import Species, LoaderState
from .errs import DuckException, DuckNameError, DuckImportError, DuckAttributeError, DuckModuleNotFoundError
from .grds import isnone, notnone, isalias, isduck, isduckspec, isiter, iscls, isclstype
from .misc import find_last_spec, prepare_types_for_instancecheck, prepare_ducktypes, prepare_isinstance_types

# %% ../nbs/09_meta.ipynb 17
class ABCSpecies(ABCMeta):
    __species__: ClassVar[Species] = Species.REG
    '''The specific "species" of the protocol. Defaults to `REG` (regular).'''
        
    def __reg__(cls) -> Self:
        '''Set the "species" to `REG` (regular).'''
        cls.__species__ = Species.REG
        return cls
    
    def __not__(cls) -> Self:
        '''Set the "species" to `NOT` (negated).'''
        cls.__species__ = Species.NOT
        return cls
    
    def __opt__(cls) -> Self:
        '''Set the "species" to `OPT` (optional).'''
        cls.__species__ = Species.OPT
        return cls
    
    def __migrate__(cls, species: Species) -> Self:
        '''Migrate the protocol to a different "species".'''
        match species:
            case Species.REG: return cls.__reg__()
            case Species.NOT: return cls.__not__()
            case Species.OPT: return cls.__opt__()
            case _: return cls # NOTE: fail silently
        return cls
    
    def __hash__(self) -> int:
        return super().__hash__()

# %% ../nbs/09_meta.ipynb 19
class ABCDuck(_ProtocolMeta, ABCSpecies, type):
    __ducktype__: ClassVar[tuple[type[T], ...]] = tuple()
    '''The specific types that the protocol is duck-typed to.''';
    
    __expected__: ClassVar[dict[str, Any]] = dict()
    '''Dictionary of attributes and their expected values to check in the object.''';
    
    __required__: ClassVar[tuple[str, ...]] = tuple()
    '''Iterable of attribute names that the object must have.''';
    
    __allequal__: ClassVar[tuple[str, ...]] = tuple(ALL_ATTRS)
    '''Iterable of attributes to check if their values are all the same in an iterable object.''';

    __extrafns__: ClassVar[dict[str, Callable[[O, T], TypeGuard[T]]]] = dict()
    '''Mapping of attribute names to guard functions for additional checks.''';
    
    __usealias__: ClassVar[bool] = True
    '''Whether or not to use the `isalias` function for instance checking or `__guard__`.''';
    
    # __initiate__, __makefunc__
    __generate__: ClassVar[Callable[P, T] | T | None] = None
    '''A factory function or default value to use to initialize a valid isstance
    of the duck type.''';
    
    __areinrow__: ClassVar[bool] = False
    
    
    def __types__(cls) -> tuple[type, ...]:
        '''Return the types that the protocol is duck-typed to.'''
        curtypes = list(cls.__ducktype__)
        unducked = prepare_ducktypes(cls)
        # print('__types__', cls, cls.__name__, curtypes, unducked, cls.__areinrow__)
        cls.__ducktype__ = unducked
        return unducked
    
    def __typesq__(cls) -> tuple[type, ...]:
        '''Return the types that the protocol is duck-typed to, including NoneType.'''
        # print('__typesq__')
        return cls.__types__() + (NoneType, )
    
    def __usetypes__(cls) -> tuple[type, ...]:
        '''Return the types that the protocol is duck-typed and to use.'''
        # print('__usetypes__')
        if cls.__species__ == Species.OPT: return cls.__typesq__()
        return cls.__types__()
    
    def __guard__(cls, inst: Any) -> TypeGuard[type[T]]: 
        '''Return a type guard for the protocol.'''
        # print('__guard__')
        return isinstance(inst, prepare_isinstance_types(cls.__types__()))
    
    def __guardq__(cls, inst: Any) -> TypeGuard[type[T]]: 
        '''Return a type guard for the protocol, including NoneType.'''
        # print('__guardq__')
        return isinstance(inst, prepare_isinstance_types(cls.__typesq__()))
    
    def __alias__(cls, inst: Any) -> TypeGuard[type[T]]:
        isfn: FunctionType = partial(isalias, 
            attrs = cls.__expected__, 
            hasattrs = cls.__required__, 
            allattrs = cls.__allequal__, 
            mapattrs = cls.__extrafns__
        )
        # print('calling alias')
        match cls.__species__:
            case Species.REG: return isfn(inst, types=cls.__types__())
            case Species.NOT: return not isfn(inst, types=cls.__types__())
            case Species.OPT: return isfn(inst, types=cls.__typesq__())
        return isalias(
            inst, 
            types = cls.__types__(), 
            guards = (cls.__guard__, ),
            attrs = cls.__expected__, 
            hasattrs = cls.__required__, 
            allattrs = cls.__allequal__, 
            mapattrs = cls.__extrafns__
        )
    
    def __instancecheck__(cls, inst: Any) -> TypeGuard[type[T]]: 
        '''Instance check for the protocol using custom `__guard__` method
        which enables duck-typing and inheritance.'''
        # print('calling __instancecheck__')
        if cls.__usealias__:  
            return cls.__alias__(inst)
        match cls.__species__:
            case Species.REG: return cls.__guard__(inst)
            case Species.NOT: return cls.__notinstance__(inst)
            case Species.OPT: return cls.__optinstance__(inst)
            case _: return cls.__guard__(inst)
        return cls.__guard__(inst)
    
    def __subclasscheck__(cls, inst: Any) -> TypeGuard[type[T]]: 
        '''Instance check for the protocol using custom `__guard__` method
        which enables duck-typing and inheritance.'''
        # print('subclass check', cls, type(cls), inst, type(inst), )
        types = cls.__usetypes__()
        isin = inst in types
        # bases = []
        # for t in types:
        #     b = list(getattr(t, '__bases__', ()))
        #     bases.extend(b)
        # bases = tuple(list(filter(isclstype, bases)))
        
        match cls.__species__:
            case Species.REG: return (isin or issubclass(inst, types))
            case Species.NOT: return not (isin or issubclass(inst, types))
            case Species.OPT: return (isin or issubclass(inst, types))
        return issubclass(inst, types)
    
    def __notinstance__(cls, inst: Any) -> TypeGuard[NotProtocol[T]]:
        '''Return a type guard for NOT the protocol.'''
        return not cls.__instancecheck__(inst)

    def __optinstance__(cls, inst: Any) -> TypeGuard[OptProtocol[T]]:
        '''Return a type guard for the protocol or NoneType.'''
        return cls.__guardq__(inst)

    def __invert__(cls) -> NotProtocol[T]:
        '''Return the NOT type guard for the protocol.'''
        return cls.__notinstance__

    def __memberguard__(cls, inst: Any) -> TypeGuard[type[T]]: 
        '''Member check for the protocol using custom `__types__` method
        which enables duck-typing and inheritance.'''
        types = cls.__types__()
        if cls.__species__ == Species.OPT: types = cls.__typesq__()
        def checkmember(inst: Any) -> TypeGuard[type[T]]:
            return bool((inst in types) ^ (cls.__species__ == Species.NOT))
        return checkmember

    def __membercheck__(cls, inst: Any) -> TypeGuard[type[T]]: 
        '''Member check for the protocol using custom `__types__` method
        which enables duck-typing and inheritance.'''
        match cls.__species__:
            case Species.REG: return inst in cls.__types__()
            case Species.NOT: return not inst in cls.__types__()
            case Species.OPT: return inst in cls.__typesq__()
            case _: return inst in cls.__types__()

    def __contains__(cls, inst: Any) -> TypeGuard[type[T]]: 
        '''Instance check for the protocol using custom `__guard__` method
        which enables duck-typing and inheritance.'''
        return cls.__membercheck__(inst)
    
    def __eq__(cls, inst: Any) -> TypeGuard[type[T]]: 
        '''Equality check for the protocol using custom `__guard__` method'''
        # print('__eq__', cls, inst)
        is_implemented = False
        try:
            is_not_implemented = (super().__eq__(inst) == NotImplemented)
            is_implemented = not (is_not_implemented == True)
        except: ...
        try: 
            if cls.__instancecheck__(inst): 
                return True
        except TypeError as inserror: ...
        try:
            if cls.__subclasscheck__(inst):
                return True
        except TypeError as suberror: ...
        try: 
            if inst in cls.__types__(): return True
        except TypeError as inerror: ...
        try: 
            if inst is cls: return True
        except TypeError as iserror: ...
        return is_implemented
    
    def __repr__(cls) -> TypeGuard[type[T]]:
        name = cls.__name__
        match cls.__species__:
            case Species.REG: return f'<duck {name}>'
            case Species.NOT: return f'<!duck {name}>'
            case Species.OPT: return f'<?duck {name}>'
            
    def __maketype__(cls, __type: T) -> tuple[T, bool]:
        '''Try and make the ducktype using a given type.'''
        if isnone(__type): return __type, False
        
        # NOTE: __type is not none so unduck duckspecs and other ducktypes
        
        # NOTE: if the item is a duckspec try to load the asset stored 
        # in `duckspec.duck` or fallback to `duckspec.default`
        ins = __type
        if cls.__instancecheck__(ins): return ins, True
        
        if isduckspec(ins):
            ins = cls.__makespec__(ins)
            if cls.__instancecheck__(ins): return ins, True
        
        # NOTE: __type is a duck type (recurse / invoke) it's 
        # own `__initiate__` until resolved.
        if isduck(ins):
            ins = cls.__makeduck__(ins)
            if cls.__instancecheck__(ins): return ins, True
            
            
        # NOTE: __type is a callable so try to invoke the instance
        if not iscls(ins) and callable(ins):
            ins = cls.__makefunc__(ins)
            if cls.__instancecheck__(ins): return ins, True
    
        # NOTE: __type is not a class or iterable of classes but an instance already, 
        # so try to copy it
        if not iscls(ins) and not isiter(ins):
            ins = cls.__makecopy__(ins)
            if cls.__instancecheck__(ins): return ins, True
        
        if callable(ins):
            ins = cls.__makefunc__(ins)
            if cls.__instancecheck__(ins): return ins, True
            
        if iscls(ins):
            if cls.__instancecheck__(ins): return ins, True
        
        if cls.__instancecheck__(ins): return ins, True
        return __type, False
    
    def __makeiter__(cls, __iter: T) -> tuple[T, bool]:
        '''Try and make the ducktype by iterating over multiple
        types and seeing if any can be initialized into a valid instance.'''
        if isnone(__iter): return __iter, False
        # NOTE: __iter is an iterable of classes so try to 
        # unduck duckspecs, other ducktypes, invoke functions, etc
        if isiter(__iter):
            for kls in __iter:
                ins, val = cls.__maketype__(kls)
                if val: return ins, True
            return __iter, False
        return __iter, False
    
    def __makecopy__(cls, __inst: T) -> T:
        '''Try and make the ducktype by copying the given instance of said type.'''
        if isnone(__inst): return __inst
        try: return copy.copy(__inst)
        except: ...
        return __inst
    
    def __makespec__(cls, __spec: T) -> T:
        '''Try and make the ducktype by importing it via the duckspec or using its default.'''
        if isnone(__spec): return __spec
        # unduckspec input
        if isduckspec(__spec):
            try: return __spec.duck
            except: return __spec.default
        return __spec
    
    def __makeduck__(cls, __duck: T) -> T:
        '''Try and make this ducktype by recursing another ducktype.'''
        if isnone(__duck): return __duck
        # unduck input
        if isduck(__duck):
            try: return __duck.__initiate__()
            except: ...
        return __duck
     
    def __makefunc__(cls, __func: Callable[P, T]) -> T:
        '''Try and make the ducktype via a callable function.'''
        if isnone(__func): return __func
        # NOTE: __func is a callable so try to invoke the instance
        if callable(__func):
            try: return __func()
            except: ... 
        return __func
    
    def __makesupr__(cls) -> T:
        '''Try to make the ducktype by utilizing it's superclass.'''
        gen = getattr(cls, __GENERATE__, None)
        dck = getattr(cls, __DUCKTYPE__, None)
        if isnone(gen) or isnone(dck):
            try: return super().__initiate__()
            except: ...
        return None

        
    def __initiate__(cls) -> type[T] | None:
        '''Make the ducktype based on either a function, type, or iterable 
        of types, returning None if the result is not an instance of this
        duck-type or upon any type of failure.
        '''
        opt = cls.__species__ == Species.OPT
        gen = getattr(cls, __GENERATE__, None)
        dck = getattr(cls, __DUCKTYPE__, None)
        # dck = sorted(dck, key = lambda x: iscls(x), reverse=True)
        sup = None
    
        # print(gen, dck)
        # NOTE: __generator__ is provided try to invoke the instance
        # via either a function, default value, or iterable of types.
        if notnone(gen):
            if isiter(gen):
                # print('gen.isiter')
                ins, val = cls.__makeiter__(gen)
                if val: 
                    try: return ins() if iscls(ins) else ins
                    except: return ins
            else: 
                # print('gen.single')
                ins, val = cls.__maketype__(gen)
                if val: 
                    try: return ins() if iscls(ins) else ins
                    except: return None if opt else ins
            

        # NOTE: __generator__ failed to find a valid instance fallback to 
        # __ducktype__ to try and initiate the instance.
        if notnone(dck):
            if isiter(dck):
                # print('dck.isiter')
                ins, val = cls.__makeiter__(dck)
                if val: 
                    try: return ins() if iscls(ins) else ins
                    except: return None if opt else ins
            else: 
                # print('dck.single')
                ins, val = cls.__maketype__(dck)
                if val: 
                    try: return ins() if iscls(ins) else ins
                    except: return None if opt else ins
        
        
        # NOTE: both __generator__ and __ducktype__ failed to invoke
        # the instance, try to use the super class __initiate__ to
        # get the default value or duck type.
        sup = cls.__makesupr__()
        if cls.__instancecheck__(sup): return sup
        return None
            
    def __call__(cls, *args, **kwargs) -> TypeGuard[type[T]]:
        '''Return the protocol class itself.'''
        gen = any(kwargs.pop(key, False) for key in CALL_INIT_ATTRS)
        noargs = (not bool(args) and not bool(kwargs))
        if gen or noargs:
            try: return cls.__initiate__()
            except: ...
            
        fn = cls.__instancecheck__
        results = [fn(arg) for arg in args]
        if len(results) == 1: return results[0]
        return tuple(results)

    def __hash__(self) -> int:
        return super().__hash__()
    
    def __new__(mcls, name, bases, namespace, **kwargs):
        '''Create a new protocol class.'''
        # print('meta.__new__\t', mcls, name, bases, namespace, kwargs)
        new = super().__new__(mcls, name, bases, namespace)
        new = mcls.__metahook__(new, name, bases, namespace, **kwargs)
        return new
        
    @classmethod
    def __metahook__(mcls, cls, name, bases, namespace, **kwargs):
        '''Hook for the protocol metaclass.'''
        # print('meta.__metahook__\t', mcls, cls, name, bases, namespace, kwargs)
        hook = getattr(cls, '__metahook__', None)
        if callable(hook): 
            adj = hook(mcls, name, bases, namespace, **kwargs)
        return cls
    
    def __init__(cls, *args, **kargs):
        ...
        # print('__init__\t', cls, args, kargs)
        cls.__areinrow__ = False
    
    def __init_subclass__(cls, *args, **kwargs):
        super().__init_subclass__()
        # print('meta.__init_subclass__\t', cls, args, kwargs)
        return 

# %% ../nbs/09_meta.ipynb 40
class DuckProtocolMeta(_ProtocolMeta, ABCSpecies, ABCMeta, type):
    __ducktype__: ClassVar[tuple[type[T], ...]] = tuple()
    '''The specific types that the protocol is duck-typed to.''';
    
    _duck: ClassVar[tuple[type[T], ...]]
    @property
    def __duck__(cls) -> tuple[type[T], ...]:
        if getattr(cls, _DUCK, None) is None:
            unducked = prepare_types_for_instancecheck(cls.__ducktype__)
            setattr(cls, _DUCK, unducked)
        return cls._duck
    
    @abstractmethod
    def __types__(cls) -> tuple[type, ...]:
        '''Return the types that the protocol is duck-typed to.'''
        return prepare_types_for_instancecheck(cls.__duck__)
        duck = list(cls._duck)
        for i, d in enumerate(duck):
            if not (isinstance(d, type) or isclass(d)):
                duck[i] = type(d)
        return tuple(duck)

    @abstractmethod
    def __typesq__(cls) -> tuple[type, ...]: 
        '''Return the types that the protocol is duck-typed to, including NoneType.'''
        return cls.__types__() + (NoneType, )
        
    @abstractmethod
    def __guard__(cls, inst: Any) -> TypeGuard[type[T]]: 
        '''Return a type guard for the protocol.'''
        return isinstance(inst, cls.__types__())
    
    @abstractmethod
    def __guardq__(cls, inst: Any) -> TypeGuard[type[T]]: 
        '''Return a type guard for the protocol, including NoneType.'''
        return isinstance(inst, cls.__typesq__())
        
    def __instancecheck__(cls, inst: Any) -> TypeGuard[type[T]]: 
        '''Instance check for the protocol using custom `__guard__` method
        which enables duck-typing and inheritance.'''
        match cls.__species__:
            case Species.REG: return cls.__guard__(inst)
            case Species.NOT: return cls.__notinstance__(inst)
            case Species.OPT: return cls.__optinstance__(inst)
        return cls.__guard__(inst)
    
    def __subclasscheck__(cls, inst: Any) -> TypeGuard[type[T]]: 
        '''Instance check for the protocol using custom `__guard__` method
        which enables duck-typing and inheritance.'''
        match cls.__species__:
            case Species.REG: return issubclass(inst, cls.__types__())
            case Species.NOT: return not issubclass(inst, cls.__types__())
            case Species.OPT: return issubclass(inst, cls.__typesq__())
        return issubclass(inst, cls.__types__())
    
    @abstractmethod
    def __notinstance__(cls, inst: Any) -> TypeGuard[NotProtocol[T]]:
        '''Return a type guard for NOT the protocol.'''
        return not cls.__instancecheck__(inst)
    
    @abstractmethod
    def __optinstance__(cls, inst: Any) -> TypeGuard[OptProtocol[T]]:
        '''Return a type guard for the protocol or NoneType.'''
        return cls.__guardq__(inst)
    
    @abstractmethod
    def __invert__(cls) -> NotProtocol[T]:
        '''Return the NOT type guard for the protocol.'''
        return cls.__notinstance__
    
    @abstractmethod
    def __membercheck__(cls, inst: Any) -> TypeGuard[type[T]]: 
        '''Member check for the protocol using custom `__types__` method
        which enables duck-typing and inheritance.'''
        match cls.__species__:
            case Species.REG: return inst in cls.__types__()
            case Species.NOT: return not inst in cls.__types__()
            case Species.OPT: return inst in cls.__typesq__()
        return inst in cls.__types__()
    
    def __contains__(cls, inst: Any) -> TypeGuard[type[T]]: 
        '''Instance check for the protocol using custom `__guard__` method
        which enables duck-typing and inheritance.'''
        return cls.__membercheck__(inst)
    
    @abstractmethod
    def __eq__(cls, inst: Any) -> TypeGuard[type[T]]: 
        '''Equality check for the protocol using custom `__guard__` method'''
        try: 
            if cls.__instancecheck__(inst): 
                return True
        except TypeError as inserror: ...
        try: 
            if cls.__subclasscheck__(inst):
                return True
        except TypeError as suberror: ...
        try: 
            if inst in cls.__types__():
                return True
        except TypeError as inerror: ...
        try: 
            if inst is cls:
                return True
        except TypeError as iserror: ...
        return False
    
    def __hash__(self) -> int:
        return super().__hash__()

# %% ../nbs/09_meta.ipynb 42
class DuckModuleMeta(ABCMeta, type):
    __duckmodule__: ClassVar[str]
    '''The name of the module to duck.''';
    
    __hush__: ClassVar[bool] = True
    '''Whether or not to raise AttributeError when an attribute is not found.''';
    
    __alts__: ClassVar[dict[str, type]] = {}
    '''Alternatives to use when an attribute is not found.''';
    
    def __instancecheck__(cls: type[T], obj: Any) -> bool:
        '''Check if the object is an instance of the class.'''
        issubbed = isinstance(obj, (ModuleType, ))
        samename = object.__getattribute__(cls, __DUCKMODULE__) == obj.__name__
        hasattrs = all(hasattr(obj, attr) for attr in MODULE_TYPE_ATTRS)
        hasprivs = all(hasattr(obj, attr) for attr in (__DUCKMODULE__, __HUSH__, __ALTS__))
        if issubbed and samename and (hasattrs or hasprivs): return True
        return False
     
    
    def __getattr__(cls: type[T], name: str) -> Any:
        hush = object.__getattribute__(cls, __HUSH__)
        try:
            if name in ANNOYING_ATTRS: 
                try:
                    # return super().__getattribute__(name)
                    raise AttributeError(f"Module '{cls.__name__}' has no attribute '{name}'.")
                except:
                    if object.__getattribute__(cls, __HUSH__): return EmptyModule
                    raise AttributeError(f'''Module '{cls.__name__}' has no attribute '{name}'.''')
            
            # base module name
            base = object.__getattribute__(cls, __DUCKMODULE__)
            # full module name
            full = f'{base}.{name}' if cls.__name__ != 'DuckModule' else name
            
            mkey = full # key to use for sys.modules
            spec: ModuleSpec | None = None # module spec
            stub: str | None = None
            
            # <base.name> not in sys modules
            if full not in sys.modules:            
                # try using importlib to load the module
                try:
                    module = import_module(full)
                    sys.modules[full] = module
                    mkey = full
                except ModuleNotFoundError: ...
                    
                # try using the last found module spec
                try:
                    spec, stub = find_last_spec(full)
                    module: ModuleType = module_from_spec(spec)
                    sys.modules[stub] = module
                    mkey = stub
                    
                except:
                    if object.__getattribute__(cls, __HUSH__): return EmptyModule
                    raise AttributeError(f'''Module '{cls.__name__}' has no attribute '{name}'.''')
                
            # <base.name> or <stub> in sys.modules
            module = sys.modules[mkey]
            
            # create a DuckModule subclass for the submodule if needed            
            # @wraps(cls, assigned=('__name__', '__qualname__', '__doc__', '__annotations__'), updated=())
            # class ducked(cls):
            #     __duckmodule__ = mkey
                
            ducked = type(mkey, (cls, ), dict(__duckmodule__=mkey, __alts__=cls.__alts__, __hush__=cls.__hush__,))
            # ducked = ducked(module.__name__, module.__doc__)
            # ducked.__dict__ = {**module.__dict__, **ducked.__dict__}
            # ducked.__dict__.update({ **module.__dict__, **ducked.__dict__,**dict(_duck=mkey, __alts__=cls.__alts__, __hush__=cls.__hush__,)})

            # Try to get the attribute from the module if it exists
            try: 
                attr = object.__getattribute__(module, name)
                if LoaderState.get(module) not in [LoaderState.NONE, LoaderState.DUCKED]:
                    sys.modules[mkey] = LoaderState.set(module, LoaderState.EXECED)
                return attr
            except AttributeError: ...
            
            # Try actually loading the module if it hasn't been loaded yet
            try: 
                sys.modules[mkey] = LoaderState.exec(module)
                return object.__getattribute__(module, name)
            except AttributeError: ...
            
            # Handle fallbacks if defined
            if hasattr(cls, __ALTS__) and name in cls.__alts__: return cls.__alts__[name]
            
            return ducked
        
        except NameError as e:
            if not hush: raise DuckNameError(*e.args)
        
        except ModuleNotFoundError as e:
            if not hush: raise DuckModuleNotFoundError(e.args, e.name, e.path)
        
        except ImportError as e:
            if not hush: raise DuckImportError(e.args, e.name, e.path)
        
        except AttributeError as e:
            if not hush: raise DuckAttributeError(e.args, e.name, e.obj)
        
        except Exception as e:
            if not hush: raise DuckException(f'NameError: {e}')
        
        return EmptyModule
    
    def __hash__(self) -> int:
        return super().__hash__()
